# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Triton implementation of Flash Attention w/ Paged KV Cache + FlashDecoding.

Compatible with A10, H100, AMD MI300X.
"""

from typing import Final

import torch
import triton
import triton.language as tl
from triton.language.extra import libdevice  # type: ignore[attr-defined]

from conch.platforms import current_platform

# FP8 representation on CUDA and ROCm
_FP8_DTYPES: Final = [torch.float8_e4m3fn, torch.float8_e4m3fnuz]


@triton.jit  # type: ignore[misc]
def _load(  # type: ignore[no-untyped-def]
    data_ptr: tl.tensor,
    use_mask,
    mask: tl.tensor,
    other,
) -> tl.tensor:
    """Load a 1D tensor with custom strides and offsets."""
    if use_mask:
        # Load with mask
        data = tl.load(data_ptr, mask=mask, other=other)
    else:
        # Load without mask
        data = tl.load(data_ptr)

    return data


@triton.jit  # type: ignore[misc]
def _store(  # type: ignore[no-untyped-def]
    data_ptr: tl.tensor,
    value: tl.tensor,
    use_mask,
    mask: tl.tensor,
) -> None:
    """Store a 1D tensor with custom strides and offsets."""
    if use_mask:
        # Store with mask
        tl.store(data_ptr, value, mask=mask)
    else:
        # Store without mask
        tl.store(data_ptr, value)


@triton.jit  # type: ignore[misc]
def _varlen_attention_compute_splits_kernel(  # noqa: PLR0913, PLR0915
    # Pointers to tensors
    output_scratchpad_ptr: tl.tensor,  # (num_kv_splits, total_num_q, num_query_heads, head_size)
    lse_scratchpad_ptr: tl.tensor,  # (num_kv_splits, total_num_q, num_query_heads)
    query_ptr: tl.tensor,  # (total_num_q, num_query_heads, head_size)
    key_cache_ptr: tl.tensor,  # (num_cache_blocks, cache_block_size, num_kv_heads, head_size)
    value_cache_ptr: tl.tensor,  # (num_cache_blocks, cache_block_size, num_kv_heads, head_size)
    block_table_ptr: tl.tensor,  # (batch_size, max_num_blocks_per_sequence)
    seq_lens_ptr: tl.tensor,  # (batch_size, )
    cu_seqlens_q_ptr: tl.tensor,  # (batch_size + 1, )
    q_scale_ptr: tl.tensor,  # (1,)
    k_scale_ptr: tl.tensor,  # (1,)
    v_scale_ptr: tl.tensor,  # (1,)
    # Scalar arguments
    scale: float,
    num_cache_blocks_per_split: int,
    softcap: float,
    # Sizes of tensors above
    head_size: int,  # output.shape[3]
    query_group_size: int,  # num_query_heads // num_kv_heads
    num_kv_splits: int,
    # Strides for tensors above
    output_scratchpad_kv_split_stride: tl.int64,  # output_scratchpad.stride(0)
    output_scratchpad_batch_stride: tl.int64,  # output_scratchpad.stride(1)
    output_scratchpad_head_stride: tl.int64,  # output_scratchpad.stride(2)
    lse_scratchpad_kv_split_stride: tl.int64,  # lse_scratchpad.stride(0)
    lse_scratchpad_batch_stride: tl.int64,  # lse_scratchpad.stride(1)
    query_batch_stride: tl.int64,  # query.stride(0)
    query_head_stride: tl.int64,  # query.stride(1)
    kv_page_stride: tl.int64,  # key_cache.stride(0), same for key and value
    kv_cache_block_stride: tl.int64,  # key_cache.stride(1), same for key and value
    kv_head_stride: tl.int64,  # key_cache.stride(2), same for key and value
    block_table_batch_stride: tl.int64,  # block_table.stride(0)
    # Constexprs
    cxpr_query_group_size_padded: tl.constexpr,  # num_query_heads // num_kv_heads
    cxpr_query_chunk_size: tl.constexpr,
    cxpr_cache_block_size: tl.constexpr,
    cxpr_head_size_padded: tl.constexpr,
    cxpr_is_softcap: tl.constexpr,
    cxpr_apply_fp8_scaling: tl.constexpr,
    cxpr_is_causal: tl.constexpr,
    cxpr_split_kv: tl.constexpr,
    cxpr_use_conditional_mask: tl.constexpr,
) -> None:
    """Varlen Attention kernel: compute attention for a split block.

    Args:
        output_scratchpad_ptr: Pointer to tensor as scratchpad for output of each cache block, shape: (num_kv_splits, total_num_q, num_query_heads, head_size).
        lse_scratchpad_ptr: Pointer to tensor as scratchpad for log-sum-exp of each cache block, shape: (num_kv_splits, total_num_q, num_query_heads).
        query_ptr: Pointer to tensor storing the query, shape: (total_num_q, num_query_heads, head_size).
        key_cache_ptr: Tensor with cached K values, shape: (num_blocks, cache_block_size, num_kv_heads, head_size).
        value_cache_ptr: Tensor with cached V values, shape: (num_blocks, cache_block_size, num_kv_heads, head_size).
        block_table_ptr: Pointer to tensor storing the mapping from batch to cache blocks, shape: (batch_size, max_num_blocks_per_sequence).
        seq_lens_ptr: Pointer to tensor holding the current sequence length for each sequence in the batch, shape: (batch_size, ).
        cu_seqlens_q_ptr: Pointer to tensor holding the cumulative sequence lengths for each sequence in the batch, shape: (batch_size, ).
        q_scale_ptr: Pointer to scalar fp8 scaling factor for q.
        k_scale_ptr: Pointer to scalar fp8 scaling factor for k.
        v_scale_ptr: Pointer to scalar fp8 scaling factor for v.
        scale: Scaling factor, 1/sqrt(head_size).
        num_cache_blocks_per_split: The maximum number of cache blocks in each split (max num each kernel will process).
        softcap: Logits softcap to apply.
        head_size: Actual head dim, not padded to power-of-two.
        query_group_size: Number of query heads in each group.
        num_kv_splits: The number of times we're splitting processing of the KV seqlen.
        output_scratchpad_kv_split_stride: Stride of the output scratchpad tensor in the 0th dimension.
        output_scratchpad_batch_stride: Stride of the output scratchpad tensor in the 1st dimension.
        output_scratchpad_head_stride: Stride of the output scratchpad tensor in the 2nd dimension.
        lse_scratchpad_kv_split_stride: Stride of the log-sum-exp scratchpad tensor in the 0th dimension.
        lse_scratchpad_batch_stride: Stride of the log-sum-exp scratchpad tensor in the 1st dimension.
        query_batch_stride: Stride of the query tensor in the 0th dimension.
        query_head_stride: Stride of the query tensor in the 1st dimension.
        kv_page_stride: Stride of the k/v tensors in the 0th dimension.
        kv_cache_block_stride: Stride of the k/v tensors in the 1st dimension.
        kv_head_stride: Stride of the k/v tensors in the 2nd dimension.
        block_table_batch_stride: Stride of the block table tensor in the 0th dimension.
        cxpr_query_group_size_padded: The number of query heads to group together (must be power-of-two!).
        cxpr_query_chunk_size: The size of the query chunks (must be power of two!).
        cxpr_cache_block_size: The size of the cache blocks (must be power of two!).
        cxpr_head_size_padded: The head size of the attention layer padded to the next power of two.
        cxpr_is_softcap: Whether or not logits softcapping will be applied.
        cxpr_apply_fp8_scaling: Whether or not to apply FP8 scaling.
        cxpr_is_causal: Whether or not to apply causal masking.
        cxpr_split_kv: Whether or not we are using FlashDecoding.
        cxpr_use_conditional_mask: Whether to use conditionally-masked loads/stores or always apply mask.
    """
    # Encode number of query and KV splits in axis=0
    split_index = tl.program_id(0)
    total_num_splits = tl.num_programs(0)

    # kv_split_index: What "KV split" of the overall data (between 1 and N KV cache blocks) is this program processing?
    kv_split_index = split_index // tl.cdiv(total_num_splits, num_kv_splits)
    # query_split_index: What "query split" of the overall data (between 1 and M chunks of the query sequence) is this program processing?
    query_split_index = split_index % tl.cdiv(total_num_splits, num_kv_splits)

    # What KV head is this program processing?
    kv_head_index = tl.program_id(1)

    # What batch does the query that this program is processing belong to?
    batch_index = tl.program_id(2)

    # Load scalar current_sequence_length for the current sequence in the batch
    # This is the KV sequence length, not the Q sequence length
    current_sequence_length = tl.load(seq_lens_ptr + batch_index)

    # The length of the current sequence will tell us how many cache blocks we need to read
    current_seq_num_cache_blocks = tl.cdiv(current_sequence_length, cxpr_cache_block_size)

    # What is the first cache block index in the set for this split
    starting_cache_block_index = kv_split_index * num_cache_blocks_per_split

    # We launch the same number of splits for all sequences in the batch, so different kernel launches will have different numbers of cache blocks
    # to process. So we may have a case where one sequence in a batch has sufficiently fewer cache blocks to process such that a kernel doesn't have
    # any blocks to process.
    if starting_cache_block_index >= current_seq_num_cache_blocks:
        return

    # Compute length of current sequence's query
    this_query_start = tl.load(cu_seqlens_q_ptr + batch_index)
    this_query_end = tl.load(cu_seqlens_q_ptr + batch_index + 1)
    this_query_length = this_query_end - this_query_start

    # If the query length is just one token, then we have a decode case
    is_pure_decode = this_query_length == 1

    # Offset for how many tokens in query correspond to previous splits for this sequence
    this_query_split_offset = query_split_index * cxpr_query_chunk_size

    # Similar to above, we launch the same number of splits for all sequences in the batch, so different kernel launches will have different
    # numbers of query tokens to process. If we've already processed all of the query tokens for this sequence, we can skip this kernel.
    if this_query_split_offset > this_query_length:
        return

    # What is the last Q token in this block?
    end_seqlen_q = this_query_split_offset + cxpr_query_chunk_size

    # Only need causal masking if enabled and this program is not processing a decode
    needs_causal_mask = cxpr_is_causal and not is_pure_decode

    # Use dtype of output as the dtype for intermediate values/calculations (except for those that must be done in FP32)
    dtype = output_scratchpad_ptr.dtype.element_ty

    # How many tokens of K/V have we already processed prior to this kernel
    beginning_seqlen_k = starting_cache_block_index * cxpr_cache_block_size

    # How many cache blocks should this program process
    num_cache_blocks_to_process = num_cache_blocks_per_split

    if needs_causal_mask:
        if end_seqlen_q <= beginning_seqlen_k:
            return

        # How many cache blocks do we need to process until seqlen_k > end_seqlen_q
        num_cache_blocks_to_process = min(
            num_cache_blocks_per_split, tl.cdiv(end_seqlen_q - beginning_seqlen_k, cxpr_cache_block_size)
        )
    else:
        # Handle if the cache blocks for this split will exceed the seqlen_k
        prev_cache_blocks = kv_split_index * num_cache_blocks_per_split
        num_cache_blocks_to_process = min(num_cache_blocks_per_split, current_seq_num_cache_blocks - prev_cache_blocks)

    # Offsets for each query vector in the split/group
    query_split_group_offsets = tl.arange(0, cxpr_query_chunk_size * cxpr_query_group_size_padded)
    # What query vector in the split?
    query_split_group_seq_offsets = this_query_split_offset + query_split_group_offsets // cxpr_query_group_size_padded
    # What query head in the group?
    query_split_group_head_offsets = (
        kv_head_index * query_group_size
    ) + query_split_group_offsets % cxpr_query_group_size_padded

    # Need to mask out if any of these tokens are out of bounds
    query_split_group_seq_mask = query_split_group_seq_offsets < this_query_length
    query_split_group_head_mask = query_split_group_head_offsets < (kv_head_index * query_group_size) + query_group_size

    # Offsets to each element of the padded-to-next-power-of-two head size
    head_offsets = tl.arange(0, cxpr_head_size_padded)
    # Mask to only read valid indices of the actual head size
    head_mask = head_offsets < head_size

    # Offsets for the queries in this block
    query_offsets = (
        this_query_start * query_batch_stride
        + query_split_group_seq_offsets[:, None] * query_batch_stride
        + query_split_group_head_offsets[:, None] * query_head_stride
        + head_offsets[None, :]
    )

    needs_query_split_mask = True
    needs_query_group_mask = True
    needs_head_mask = True
    needs_query_mask = True

    if cxpr_use_conditional_mask:
        needs_query_split_mask = end_seqlen_q > this_query_length
        needs_query_group_mask = query_group_size != cxpr_query_group_size_padded
        needs_head_mask = head_size != cxpr_head_size_padded
        needs_query_mask = (needs_query_split_mask or needs_query_group_mask) or needs_head_mask

    # Mask out query elements that are just for padding
    query_mask = query_split_group_seq_mask[:, None] & query_split_group_head_mask[:, None] & head_mask[None, :]

    # Load queries
    if cxpr_use_conditional_mask:
        query = _load(query_ptr + query_offsets, use_mask=needs_query_mask, mask=query_mask, other=0.0)
    else:
        query = tl.load(query_ptr + query_offsets, mask=query_mask, other=0.0, eviction_policy="evict_last")

    if cxpr_apply_fp8_scaling:
        q_scale = tl.load(q_scale_ptr)
        query = (query * q_scale).to(dtype)

    # Index/offset for the current kv_head in the key_cache and value_cache
    kv_head_index_offset = kv_head_index * kv_head_stride

    # Pointer arithmetic to get to the entry in the block_table for the current batch_index
    current_block_table_offset = batch_index * block_table_batch_stride
    current_block_table_ptr = block_table_ptr + current_block_table_offset

    # Scratchpad for output from this group of cache blocks
    output = tl.zeros([cxpr_query_chunk_size * cxpr_query_group_size_padded, cxpr_head_size_padded], dtype=dtype)
    # Keep running max of softmax numerator (scale * Q * K)
    m_i = tl.full([cxpr_query_chunk_size * cxpr_query_group_size_padded], -float("inf"), dtype=dtype)
    # Keep running denominator of softmax
    l_i = tl.full([cxpr_query_chunk_size * cxpr_query_group_size_padded], 0.0, dtype=dtype)

    # Iterate through the cache blocks that this kernel is assigned to
    for cache_block_index in range(
        starting_cache_block_index, starting_cache_block_index + num_cache_blocks_to_process
    ):
        # Calculate number of entries in this cache block (will be a value between 1 and cache_block_size)
        num_entries_in_cache_block = min(
            current_sequence_length - (cache_block_index * cxpr_cache_block_size), cxpr_cache_block_size
        )

        # Offsets/mask for each entry in this cache block
        cache_block_offsets = tl.arange(0, cxpr_cache_block_size)
        cache_block_mask = cache_block_offsets < num_entries_in_cache_block

        needs_cache_block_mask = True
        needs_qk_mask = True

        if cxpr_use_conditional_mask:
            needs_cache_block_mask = num_entries_in_cache_block != cxpr_cache_block_size
            needs_qk_mask = (needs_query_split_mask or needs_query_group_mask) or needs_cache_block_mask

        # Offset from the block_table row for the current batch by the number of cache blocks
        current_cache_block_number_ptr = current_block_table_ptr + cache_block_index
        physical_cache_block_number = tl.load(current_cache_block_number_ptr)

        # Calculate address of current cache block
        kv_cache_block_index_offset = physical_cache_block_number * kv_page_stride

        # Load the key block as (cxpr_head_size_padded, cache_block_size)
        # Note: we're loading it transposed here
        key_block_offsets = (
            cache_block_offsets[None, :] * kv_cache_block_stride + kv_head_index_offset + head_offsets[:, None]
        )

        key_block_mask = head_mask[:, None] & cache_block_mask[None, :]

        if cxpr_use_conditional_mask:
            key_block = _load(
                key_cache_ptr + kv_cache_block_index_offset + key_block_offsets,
                use_mask=(needs_cache_block_mask or needs_head_mask),
                mask=key_block_mask,
                other=0.0,
            )
        else:
            key_block = tl.load(
                key_cache_ptr + kv_cache_block_index_offset + key_block_offsets,
                mask=key_block_mask,
                other=0.0,
            )

        if cxpr_apply_fp8_scaling:
            # Dequantize (multiply by scale factor)
            k_scale = tl.load(k_scale_ptr)
            key_block = (key_block * k_scale).to(dtype)

        # Multiply query vector by key matrix for this cache block (and apply scaling factor)
        # query.shape -> (query_chunk_size * query_group_size, head_size)
        # key_block.shape -> (head_size, cache_block_size)
        # qk.shape -> (query_chunk_size * query_group_size, cache_block_size)
        qk = scale * tl.dot(query, key_block)

        # Need to mask out any elements that represent unused cache block entries or padding elements
        qk_mask = query_split_group_seq_mask[:, None] & query_split_group_head_mask[:, None] & cache_block_mask[None, :]

        if needs_causal_mask:
            effective_seqlen_k_offsets = cache_block_index * cxpr_cache_block_size + cache_block_offsets
            causal_mask = query_split_group_seq_offsets[:, None] >= effective_seqlen_k_offsets[None, :]
            qk_mask = qk_mask & causal_mask

        if needs_qk_mask or needs_causal_mask:
            # Set masked out elements to -inf
            qk = tl.where(qk_mask, qk, -float("inf"))

        # Handle softcapping
        if cxpr_is_softcap:
            # tanh can only accept fp32 or fp64 arguments
            qk = softcap * libdevice.tanh(qk / softcap)

        # Reduce maximum between running max and the max of (scale * Q * K) for this cache block
        m_ij = tl.maximum(m_i, tl.max(qk, axis=1))

        # Calculate numerator of softmax for this cache block
        p = tl.exp(qk - m_ij[:, None]).to(dtype)
        # Need to mask out any elements that represent unused cache block entries or padding elements
        p = tl.where(qk_mask, p, 0.0).to(dtype)

        # Calculate sum of softmax numerator for this cache block
        l_ij = tl.sum(p, axis=1).to(dtype)

        # Calculate correction factor for this cache block
        alpha = tl.exp(m_i - m_ij).to(dtype)

        # Apply scaling factor to running output
        output *= alpha[:, None]

        # Load the value block as (cache_block_size, cxpr_head_size_padded)
        value_block_offsets = (
            cache_block_offsets[:, None] * kv_cache_block_stride + kv_head_index_offset + head_offsets[None, :]
        )

        value_block_mask = cache_block_mask[:, None] & head_mask[None, :]

        if cxpr_use_conditional_mask:
            value_block = _load(
                value_cache_ptr + kv_cache_block_index_offset + value_block_offsets,
                use_mask=(needs_cache_block_mask or needs_head_mask),
                mask=value_block_mask,
                other=0.0,
            )
        else:
            value_block = tl.load(
                value_cache_ptr + kv_cache_block_index_offset + value_block_offsets,
                mask=value_block_mask,
                other=0.0,
            )

        if cxpr_apply_fp8_scaling:
            # Dequantize (multiply by scale factor)
            v_scale = tl.load(v_scale_ptr)
            value_block = (value_block * v_scale).to(dtype)

        # Multiply softmax probabilities by value matrix for this cache block
        # p.shape -> (query_chunk_size * query_group_size, cache_block_size)
        # value_block.shape -> (cache_block_size, head_size)
        # output.shape -> (query_chunk_size * query_group_size, head_size)
        output += tl.dot(p, value_block).to(dtype)

        # Update running max
        m_i = m_ij.to(dtype)
        # Update running denominator
        l_i = l_i * alpha + l_ij

    # Apply correction for denominator of these cache blocks
    output /= l_i[:, None]

    # Calculate offsets to store the output for this query split/query group
    # 2D block of shape (query_chunk_size * query_group_size, head_size_padded)
    output_scratch_offsets = (
        kv_split_index * output_scratchpad_kv_split_stride
        + this_query_start * output_scratchpad_batch_stride
        + query_split_group_seq_offsets[:, None] * output_scratchpad_batch_stride
        + query_split_group_head_offsets[:, None] * output_scratchpad_head_stride
        + head_offsets[None, :]
    )

    # Store output scratchpad results
    if cxpr_use_conditional_mask:
        _store(
            output_scratchpad_ptr + output_scratch_offsets,
            output,
            use_mask=needs_query_mask,
            mask=query_mask,
        )
    else:
        tl.store(
            output_scratchpad_ptr + output_scratch_offsets,
            output,
            mask=query_mask,
        )

    if cxpr_split_kv:
        # Calculate scratchpad log(sum(exp))
        # Note: log() only accepts fp32/fp64 arguments
        lse = m_i + tl.log(l_i.to(tl.float32)).to(dtype)

        # Calculate offsets to store log-sum-exp for this query split/query group
        # 1D block of shape (query_chunk_size * query_group_size,)
        lse_scratch_offsets = (
            kv_split_index * lse_scratchpad_kv_split_stride
            + this_query_start * lse_scratchpad_batch_stride
            + query_split_group_seq_offsets * lse_scratchpad_batch_stride
            + query_split_group_head_offsets
        )

        lse_mask = query_split_group_seq_mask & query_split_group_head_mask

        # Store lse scratchpad results
        if cxpr_use_conditional_mask:
            _store(
                lse_scratchpad_ptr + lse_scratch_offsets,
                lse,
                use_mask=(needs_query_split_mask or needs_query_group_mask),
                mask=lse_mask,
            )
        else:
            tl.store(
                lse_scratchpad_ptr + lse_scratch_offsets,
                lse,
                mask=lse_mask,
            )


@triton.jit  # type: ignore[misc]
def _varlen_attention_reduce_splits_kernel(  # noqa: PLR0913
    # Pointers to tensors
    output_ptr: tl.tensor,  # (total_num_q, num_query_heads, head_size)
    output_scratchpad_ptr: tl.tensor,  # (num_kv_splits, total_num_q, num_query_heads, head_size)
    lse_scratchpad_ptr: tl.tensor,  # (num_kv_splits, total_num_q, num_query_heads)
    seq_lens_ptr: tl.tensor,  # (batch_size, )
    # Scalars
    num_cache_blocks_per_split: int,
    # Sizes of tensors above
    head_size: int,  # output.shape[2]
    batch_size: int,
    # Strides for tensors above
    output_batch_stride: tl.int64,  # output.stride(0)
    output_head_stride: tl.int64,  # output.stride(1)
    output_scratchpad_kv_split_stride: tl.int64,  # output_scratchpad.stride(0)
    output_scratchpad_batch_stride: tl.int64,  # output_scratchpad.stride(1)
    output_scratchpad_head_stride: tl.int64,  # output_scratchpad.stride(2)
    lse_scratchpad_kv_split_stride: tl.int64,  # lse_scratchpad.stride(0)
    lse_scratchpad_batch_stride: tl.int64,  # lse_scratchpad.stride(1)
    # Constexprs
    cxpr_cache_block_size: tl.constexpr,
    cxpr_head_size_padded: tl.constexpr,
    cxpr_use_conditional_mask: tl.constexpr,
) -> None:
    """Varlen Attention kernel: reduce results across all splits.

    Args:
        output_ptr: Pointer to tensor for final output, shape: (total_num_q, num_query_heads, head_size).
        output_scratchpad_ptr: Pointer to tensor as scratchpad for output of each cache block, shape: (num_kv_splits, total_num_q, num_query_heads, head_size).
        lse_scratchpad_ptr: Pointer to tensor as scratchpad for log-sum-exp of each cache block, shape: (num_kv_splits, total_num_q, num_query_heads).
        seq_lens_ptr: Pointer to tensor holding the current sequence length for each sequence in the batch, shape: (batch_size, ).
        num_cache_blocks_per_split: The maximum number of cache blocks each split will process.
        head_size: Actual head dim, not padded to power-of-two.
        batch_size: Number of sequences in the batch.
        output_batch_stride: Stride of the output tensor in the 0th dimension.
        output_head_stride: Stride of the output tensor in the 1st dimension.
        output_scratchpad_kv_split_stride: Stride of the output scratchpad tensor in the 0th dimension.
        output_scratchpad_batch_stride: Stride of the output scratchpad tensor in the 1st dimension.
        output_scratchpad_head_stride: Stride of the output scratchpad tensor in the 2nd dimension.
        lse_scratchpad_kv_split_stride: Stride of the log-sum-exp scratchpad tensor in the 0th dimension.
        lse_scratchpad_batch_stride: Stride of the log-sum-exp scratchpad tensor in the 1st dimension.
        cxpr_cache_block_size: The size of the cache blocks (must be power of two!), as constexpr so that we can use for reshaping tensors.
        cxpr_head_size_padded: The head size of the attention layer padded to the next power of two.
        cxpr_use_conditional_mask: Whether to use conditionally-masked loads/stores or always apply mask.
    """
    # What query head is this program processing?
    query_head_index = tl.program_id(0)
    # What batch is this program processing?
    batch_index = tl.program_id(1)

    # Datatype for intermediates/output
    dtype = output_scratchpad_ptr.dtype.element_ty

    # Accumulator for the output of this batch/head
    output = tl.zeros([cxpr_head_size_padded], dtype=dtype)
    # Running max of block lse
    m_i = tl.full([], -float("inf"), dtype=dtype)
    # Running final scale factor
    l_i = tl.full([], 0.0, dtype=dtype)

    # Load scalar current_sequence_length for the current batch
    current_sequence_length = tl.load(seq_lens_ptr + batch_index)

    # The length of the current sequence will tell us how many cache blocks we need to read
    current_seq_num_cache_blocks = tl.cdiv(current_sequence_length, cxpr_cache_block_size)

    # How many KV splits do we need to process
    num_kv_splits_this_seq = tl.cdiv(current_seq_num_cache_blocks, num_cache_blocks_per_split)

    # Offsets to each element of the padded-to-next-power-of-two head size
    head_offsets = tl.arange(0, cxpr_head_size_padded)
    # Mask to only read valid indices of the actual head size
    head_mask = head_offsets < head_size

    needs_head_mask = True

    if cxpr_use_conditional_mask:
        # Only enable masking if its necessary
        needs_head_mask = head_size != cxpr_head_size_padded

    # Iterate through every cache block for the current sequence
    for kv_split_index in range(num_kv_splits_this_seq):
        # Calculate offsets to load the scratch for this head/batch/split
        # Block of shape (cxpr_head_size_padded,)
        output_scratchpad_offsets = (
            kv_split_index * output_scratchpad_kv_split_stride
            + batch_index * output_scratchpad_batch_stride
            + query_head_index * output_scratchpad_head_stride
            + head_offsets
        )

        # Load output for this cache block, shape -> (cxpr_head_size_padded,)
        if cxpr_use_conditional_mask:
            block_output = _load(
                output_scratchpad_ptr + output_scratchpad_offsets,
                use_mask=needs_head_mask,
                mask=head_mask,
                other=0.0,
            )
        else:
            block_output = tl.load(
                output_scratchpad_ptr + output_scratchpad_offsets,
                mask=head_mask,
                other=0.0,
            )

        # Calculate offsets to load log-sum-exp for this head/batch/cache block
        lse_scratchpad_offsets = (
            kv_split_index * lse_scratchpad_kv_split_stride
            + batch_index * lse_scratchpad_batch_stride
            + query_head_index
        )

        # Load log-sum-exp for this cache block, shape -> scalar
        block_lse = tl.load(lse_scratchpad_ptr + lse_scratchpad_offsets)

        # Reduce running max lse
        m_ij = tl.maximum(m_i, block_lse)

        # Calculate correction factor from previous cache blocks
        alpha = tl.exp((m_i - m_ij).to(tl.float32)).to(dtype)

        # Apply correction factor
        output *= alpha

        # Calculate correction factor from this cache block
        # Note: exp() only accepts fp32/fp64 arguments
        beta = tl.exp((block_lse - m_ij).to(tl.float32)).to(dtype)

        # Apply second correction factor and accumulate running output
        output += beta * block_output

        # Update running max
        m_i = m_ij.to(dtype)
        # Update running final scale factor
        l_i = l_i * alpha + beta

    # Apply final correction to output
    output /= l_i

    # Calculate offsets to store the output for this head/batch
    output_offsets = batch_index * output_batch_stride + query_head_index * output_head_stride + head_offsets

    # Store final result
    if cxpr_use_conditional_mask:
        _store(
            output_ptr + output_offsets,
            output,
            use_mask=needs_head_mask,
            mask=head_mask,
        )
    else:
        tl.store(
            output_ptr + output_offsets,
            output,
            mask=head_mask,
        )


def _do_conditionally_masked_loads_and_stores() -> bool:
    """Determine whether we should do conditionally-masked loads and stores or always use a mask.
    On some platforms (H100) its faster to always use the mask than to conditionally apply it.
    On MI300X, its much faster to conditionallly apply the mask.
    """
    # TODO(jmanning): Customize this per-platform
    return "H100" not in current_platform.get_device_name()


def _get_block_size() -> int:
    """Get block size for tuning purposes."""
    # Note: in the future we could specify different settings here based on the platform or autotune,
    # but for simplicity right now we'll hardcode this block size because it gives good performance
    # on A10, H100, and MI300X.
    return 32


def _get_tuned_sizes(head_size_padded: int, query_group_size_padded: int, max_seqlen_q: int) -> tuple[int, int]:
    """Get tuned sizes for current device."""
    block_size = _get_block_size()

    # If the head size grows too large then we want to limit how many queries we're chunking together
    small_head_size_limit: Final = 128
    if head_size_padded > small_head_size_limit:
        block_size = block_size // 2

    # The "query chunk size" represents the number of queries that each kernel will process at a time.
    query_chunk_size = max(1, block_size // query_group_size_padded) if max_seqlen_q > 1 else 1

    # If we have all decodes, we need to make sure that we have at least a block size of 16 for tl.dot
    query_group_size_padded = query_group_size_padded if max_seqlen_q > 1 else max(16, query_group_size_padded)

    return query_chunk_size, query_group_size_padded


def varlen_attention_launcher(  # noqa: PLR0913
    output: torch.Tensor,
    query: torch.Tensor,
    key_cache: torch.Tensor,
    value_cache: torch.Tensor,
    cu_seqlens_q: torch.Tensor,
    max_seqlen_q: int,
    seq_lens: torch.Tensor,
    max_seqlen_k: int,
    block_table: torch.Tensor,
    output_scratchpad: torch.Tensor | None = None,
    lse_scratchpad: torch.Tensor | None = None,
    causal: bool = False,
    scale: float | None = None,
    softcap: float = 0.0,
    kv_cache_dtype: str = "auto",
    q_scale: torch.Tensor | None = None,
    k_scale: torch.Tensor | None = None,
    v_scale: torch.Tensor | None = None,
    strict: bool = False,
) -> None:
    """Varlen Attention kernel launcher.

    Args:
        output: Tensor to write the output of the attention calculation, shape: (total_num_q, num_heads, head_size).
        query: Query tensor, shape: (total_num_q, num_heads, head_size).
        key_cache: Tensor with cached K values, shape: (num_blocks, cache_block_size, num_kv_heads, head_size).
        value_cache: Tensor with cached V values, shape: (num_blocks, cache_block_size, num_kv_heads, head_size).
        cu_seqlens_q: Tensor with the cumulative query sequence lengths for each index in the batch, shape: (batch_size + 1, ).
        max_seqlen_q: Maximum sequence length of the query.
        seq_lens: Tensor with the sequence length of each index in the batch, shape: (batch_size, ).
        max_seqlen_k: Maximum sequence length of the key/value.
        block_table: Tensor storing the mapping from batch to cache blocks, shape: (batch_size, max_num_blocks_per_sequence).
        output_scratchpad: Tensor used as scratchpad to share cache block outputs between two stages, shape: (num_kv_splits, total_num_q, num_query_heads, head_size)
        lse_scratchpad: Tensor used as scratchpad to share cache block log-sum-exp between two stages, shape: (num_kv_splits, total_num_q, num_query_heads)
        causal: Whether or not to apply causal masking.
        scale: Scaling factor, 1/sqrt(head_size).
        softcap: Softcap value to apply to logits.
        kv_cache_dtype: Data type of the key/value cache.
        q_scale: Scaling factor for Q values.
        k_scale: Scaling factor for K values.
        v_scale: Scaling factor for V values.
        strict: (Optional) Whether to enable strict checking of input sizes.
    """
    if strict:
        assert query.shape == output.shape
        assert key_cache.shape == value_cache.shape
        assert key_cache.stride(0) == value_cache.stride(0)
        assert key_cache.stride(1) == value_cache.stride(1)
        assert key_cache.stride(2) == value_cache.stride(2)
        assert key_cache.stride(3) == value_cache.stride(3)
        assert key_cache.stride(3) == 1
        assert softcap >= 0.0

        expected_output_dtype = torch.bfloat16 if query.dtype in _FP8_DTYPES else query.dtype
        allowed_in_out_dtypes = [torch.float32, torch.float16, torch.bfloat16] + _FP8_DTYPES
        assert query.dtype in allowed_in_out_dtypes
        assert output.dtype == expected_output_dtype

        if output_scratchpad is not None or lse_scratchpad is not None:
            assert output_scratchpad is not None
            assert lse_scratchpad is not None
            assert output_scratchpad.dtype == expected_output_dtype
            assert lse_scratchpad.dtype == expected_output_dtype
            assert output_scratchpad.size(1) == lse_scratchpad.size(1)

    # Perform unchecked size accesses, assume has already been checked
    total_num_q, num_query_heads, head_size = output.shape
    num_cache_blocks, cache_block_size, num_kv_heads, _ = key_cache.shape
    batch_size, max_num_blocks_per_sequence = block_table.shape
    max_num_kv_splits = output_scratchpad.size(0) if output_scratchpad is not None else 1

    if strict:
        assert cache_block_size == triton.next_power_of_2(cache_block_size), "Cache block size must be a power of two!"

    # Need sizes to be constexpr in order to reshape tensors in kernel
    cxpr_cache_block_size: tl.constexpr = cache_block_size
    cxpr_head_size_padded: tl.constexpr = triton.next_power_of_2(head_size)

    # For parity with Dao Flash Attention, softcap == 0.0 means no softcapping
    cxpr_is_softcap: tl.constexpr = softcap > 0.0

    cxpr_apply_fp8_scaling: tl.constexpr = "fp8" in kv_cache_dtype or query.dtype in _FP8_DTYPES
    cxpr_use_conditional_mask: tl.constexpr = _do_conditionally_masked_loads_and_stores()

    if strict and cxpr_apply_fp8_scaling:
        assert q_scale is not None
        assert k_scale is not None
        assert v_scale is not None
        assert q_scale.numel() == 1
        assert k_scale.numel() == 1
        assert v_scale.numel() == 1

    # How many query heads correspond to the same KV head?
    query_group_size = num_query_heads // num_kv_heads
    query_group_size_padded = triton.next_power_of_2(query_group_size)

    # What is the maximum number of stage 1 kernels to launch per batch/head?
    # Each kernel processes up to {cache_block_size} tokens at a time (for each split that a stage1 kernel processes), so we
    # can process a sequence up to {max_num_kv_splits * cache_block_size} tokens before a stage 1 kernel will process multiple
    # cache blocks. This helps to reduce the overhead of kernel launches / split reduction for long sequences.
    # Note: we may need to tune this value for a given HW platform.
    num_kv_splits = min(max_num_blocks_per_sequence, max_num_kv_splits)

    # Different platforms may require different sizes.
    query_chunk_size, query_group_size_padded = _get_tuned_sizes(
        cxpr_head_size_padded, query_group_size_padded, max_seqlen_q
    )

    # Use the maximum Q sequence length to determine what is the max number of query splits any sequence will need
    num_query_splits = triton.cdiv(max_seqlen_q, query_chunk_size)

    # How many cache blocks will each kernel process?
    num_cache_blocks_per_split = triton.cdiv(max_num_blocks_per_sequence, num_kv_splits)

    # Use default scaling factor if not provided
    if scale is None:
        scale = float(1.0 / (head_size**0.5))

    output_scratchpad_kv_split_stride = 0
    output_scratchpad_batch_stride = output.stride(0)
    output_scratchpad_head_stride = output.stride(1)
    lse_scratchpad_kv_split_stride = 0
    lse_scratchpad_batch_stride = 0

    if num_kv_splits > 1:
        assert output_scratchpad is not None
        assert lse_scratchpad is not None

        output_scratchpad_kv_split_stride = output_scratchpad.stride(0)
        output_scratchpad_batch_stride = output_scratchpad.stride(1)
        output_scratchpad_head_stride = output_scratchpad.stride(2)
        lse_scratchpad_kv_split_stride = lse_scratchpad.stride(0)
        lse_scratchpad_batch_stride = lse_scratchpad.stride(1)

    # For computing attention for split block (stage 1): parallelize over query splits, KV splits, KV heads, and batches.
    stage1_grid = (num_query_splits * num_kv_splits, num_kv_heads, batch_size)

    # Launch stage 1 kernel
    _varlen_attention_compute_splits_kernel[stage1_grid](
        # Relevant tensors
        output_scratchpad_ptr=output_scratchpad if num_kv_splits > 1 else output,
        lse_scratchpad_ptr=lse_scratchpad,
        query_ptr=query,
        key_cache_ptr=key_cache,
        value_cache_ptr=value_cache,
        block_table_ptr=block_table,
        seq_lens_ptr=seq_lens,
        cu_seqlens_q_ptr=cu_seqlens_q,
        q_scale_ptr=q_scale,
        k_scale_ptr=k_scale,
        v_scale_ptr=v_scale,
        # Scalars
        scale=scale,
        num_cache_blocks_per_split=num_cache_blocks_per_split,
        softcap=softcap,
        head_size=head_size,
        query_group_size=query_group_size,
        num_kv_splits=num_kv_splits,
        # Strides of relevant tensors
        output_scratchpad_kv_split_stride=output_scratchpad_kv_split_stride,
        output_scratchpad_batch_stride=output_scratchpad_batch_stride,
        output_scratchpad_head_stride=output_scratchpad_head_stride,
        lse_scratchpad_kv_split_stride=lse_scratchpad_kv_split_stride,
        lse_scratchpad_batch_stride=lse_scratchpad_batch_stride,
        query_batch_stride=query.stride(0),
        query_head_stride=query.stride(1),
        kv_page_stride=key_cache.stride(0),
        kv_cache_block_stride=key_cache.stride(1),
        kv_head_stride=key_cache.stride(2),
        block_table_batch_stride=block_table.stride(0),
        # Constexpr sizes
        cxpr_query_group_size_padded=query_group_size_padded,
        cxpr_query_chunk_size=query_chunk_size,
        cxpr_cache_block_size=cxpr_cache_block_size,
        cxpr_head_size_padded=cxpr_head_size_padded,
        cxpr_is_softcap=cxpr_is_softcap,
        cxpr_apply_fp8_scaling=cxpr_apply_fp8_scaling,
        cxpr_is_causal=causal,
        cxpr_split_kv=(num_kv_splits > 1),
        cxpr_use_conditional_mask=cxpr_use_conditional_mask,
    )

    if num_kv_splits > 1:
        assert output_scratchpad is not None
        assert lse_scratchpad is not None

        # For reducing over splits (stage 2): parallelize over query heads and batches
        stage2_grid = (num_query_heads, batch_size)

        # Launch stage 2 kernel
        _varlen_attention_reduce_splits_kernel[stage2_grid](
            # Relevant tensors
            output_ptr=output,
            output_scratchpad_ptr=output_scratchpad,
            lse_scratchpad_ptr=lse_scratchpad,
            seq_lens_ptr=seq_lens,
            # Scalars
            num_cache_blocks_per_split=num_cache_blocks_per_split,
            head_size=head_size,
            batch_size=batch_size,
            # Strides of relevant tensors
            output_batch_stride=output.stride(0),
            output_head_stride=output.stride(1),
            output_scratchpad_kv_split_stride=output_scratchpad_kv_split_stride,
            output_scratchpad_batch_stride=output_scratchpad_batch_stride,
            output_scratchpad_head_stride=output_scratchpad_head_stride,
            lse_scratchpad_kv_split_stride=lse_scratchpad_kv_split_stride,
            lse_scratchpad_batch_stride=lse_scratchpad_batch_stride,
            # Constexpr sizes
            cxpr_cache_block_size=cxpr_cache_block_size,
            cxpr_head_size_padded=cxpr_head_size_padded,
            cxpr_use_conditional_mask=cxpr_use_conditional_mask,
        )



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Bitsandbytes quantize blockwise benchmark."""

import sys
from typing import Final

import click
import torch

from conch import envs
from conch.ops.quantization.bitsandbytes.functional import quantize_4bit as conch_quantize_4bit
from conch.platforms import current_platform
from conch.third_party.vllm.utils import seed_everything
from conch.utils.benchmark import BenchmarkMetadata, benchmark_it


def _to_torch_dtype(dtype_str: str) -> torch.dtype:
    """Map click arg for dtype to torch type."""
    if dtype_str == "uint8":
        return torch.uint8
    if dtype_str == "bf16":
        return torch.bfloat16
    if dtype_str == "fp16":
        return torch.float16
    if dtype_str == "fp32":
        return torch.float32

    error_msg = f"Unrecognized data type: '{dtype_str}'"
    raise ValueError(error_msg)


@click.command()
@click.option(
    "--blocksize",
    required=True,
    type=int,
    default=64,
    help="Size of quantized blocks",
)
@click.option(
    "--size-multiplier",
    required=True,
    type=float,
    default=458752,
    help="How many quantized blocks are in the input tensor",
)
@click.option(
    "--quant-type",
    required=True,
    type=click.Choice(["nf4", "fp4"]),
    default="nf4",
    help="Data type of quantized output",
)
@click.option(
    "--dequant-dtype",
    required=True,
    type=click.Choice(["fp16", "bf16", "fp32"]),
    default="bf16",
    help="Data type before quantization",
)
@click.option(
    "--quant-storage-dtype",
    required=True,
    type=click.Choice(["uint8", "fp16", "bf16", "fp32"]),
    default="uint8",
    help="Data type for storing quantized results",
)
@click.option(
    "--compress-statistics",
    is_flag=True,
    help="Flag for double-quantization",
)
@click.option(
    "--enable-bnb",
    is_flag=True,
    default=envs.CONCH_BENCH_ENABLE_ALL_REF,
    help="Flag to enable BNB reference impl",
)
@click.option(
    "--iteration-time-ms",
    required=False,
    type=int,
    default=10000,
    help="Time in milliseconds to run benchmark",
)
@click.option(
    "--warmup-time-ms",
    required=False,
    type=int,
    default=1000,
    help="Time in milliseconds to warmup before recording times",
)
@click.option(
    "--verbose",
    is_flag=True,
    help="Flag for printing verbose output",
)
@click.option(
    "--gpu",
    required=False,
    type=str,
    default=current_platform.device,
    help="Device to run on",
)
@click.option(
    "--csv",
    is_flag=True,
    help="Flag for printing results in CSV format",
)
def main(  # noqa: PLR0913
    blocksize: int,
    size_multiplier: float,
    quant_type: str,
    dequant_dtype: str,
    quant_storage_dtype: str,
    compress_statistics: bool,
    enable_bnb: bool,
    iteration_time_ms: int,
    warmup_time_ms: int,
    verbose: bool,
    gpu: str,
    csv: bool,
) -> None:
    """Benchmark Conch blockwise quantization kernel.

    Args:
        blocksize: Size of quantized blocks.
        size_multiplier: How many quantized blocks are in the input tensor.
        quant_type: Data type of quantized output.
        dequant_dtype: Data type before quantization.
        quant_storage_dtype: Data type for storing quantized results.
        compress_statistics: Flag for double-quantization.
        enable_bnb: Flag to enable bitsandbytes reference implementation.
        iteration_time_ms: Time in milliseconds to run benchmark.
        warmup_time_ms: Time in milliseconds to warmup before recording times.
        verbose: Flag to indicate whether or not to print verbose output.
        gpu: Which gpu to run on.
        csv: Flag to indicate whether or not to print results in CSV format.
    """
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(gpu)
    torch.set_default_device(device)

    dtype: Final = _to_torch_dtype(dequant_dtype)
    quant_storage: Final = _to_torch_dtype(quant_storage_dtype)

    input_size = int(blocksize * size_multiplier)
    x = torch.randn((input_size,), device=device, dtype=dtype)

    metadata = BenchmarkMetadata(
        platform=current_platform.name(),
        params={
            "blocksize": blocksize,
            "size_multiplier": size_multiplier,
            "quant_type": quant_type,
            "dequant_dtype": dequant_dtype,
            "quant_storage_dtype": quant_storage_dtype,
            "compress_statistics": compress_statistics,
        },
    )

    conch_output, conch_state = conch_quantize_4bit(
        x,
        absmax=None,
        out=None,
        blocksize=blocksize,
        compress_statistics=compress_statistics,
        quant_type=quant_type,
        quant_storage=quant_storage,
    )

    if enable_bnb:
        if not envs.CONCH_ENABLE_BNB:
            error_msg = "bitsandbytes must be installed and enabled via CONCH_ENABLE_BNB=1"
            raise NotImplementedError(error_msg)

        from bitsandbytes.functional import quantize_4bit as bnb_quantize_4bit

        bnb_output, bnb_state = bnb_quantize_4bit(
            x,
            absmax=None,
            out=None,
            blocksize=blocksize,
            compress_statistics=compress_statistics,
            quant_type=quant_type,
            quant_storage=quant_storage,
        )

        if not torch.allclose(conch_output, bnb_output):
            print("WARNING: Bitsandbytes and Conch results differ!", file=sys.stderr)
            print(f"Output max diff: {(bnb_output - conch_output).abs().max().item()}", file=sys.stderr)

            if verbose:
                print(f"Torch output: {bnb_output}", file=sys.stderr)
                print(f"Conch output: {conch_output}", file=sys.stderr)
        else:
            print("Results matched :)", file=sys.stderr)

        baseline_result = benchmark_it(
            lambda: bnb_quantize_4bit(
                x,
                absmax=None,
                out=None,
                blocksize=blocksize,
                compress_statistics=compress_statistics,
                quant_type=quant_type,
                quant_storage=quant_storage,
            ),
            tag="Baseline",
            metadata=metadata,
            iteration_time_ms=iteration_time_ms,
            warmup_time_ms=warmup_time_ms,
        )
    else:
        print("Skipping checking vs. reference bitsandbytes implementation...", file=sys.stderr)
        baseline_result = None

    conch_result = benchmark_it(
        lambda: conch_quantize_4bit(
            x,
            absmax=None,
            out=None,
            blocksize=blocksize,
            compress_statistics=compress_statistics,
            quant_type=quant_type,
            quant_storage=quant_storage,
        ),
        tag="Conch",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    # Print results
    conch_result.print_parameters(csv=csv)
    conch_result.print_results(csv=csv)
    if baseline_result is not None:
        baseline_result.print_results(csv=csv)


if __name__ == "__main__":
    main()



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""GEMM kernel supporting mixed-precision (e.g. w{1|2|4|8|16}, a{16|32}) and scaled matrix multiplications (fp8/int8 with scaling factors to different output dtype).

Heavily based on [GemLite](https://github.com/mobiusml/gemlite).
Core algorithm is the same, but we add a new frontend for various use-cases (mixed-precision and scaled).
"""

from dataclasses import dataclass
from enum import Enum

import torch
import triton
import triton.language as tl

from conch.platforms import current_platform


class DType(Enum):
    """Data types for mapping between Triton/Torch."""

    FP32 = 0
    FP16 = 1
    BF16 = 2
    FP8 = 3
    INT8 = 4
    UINT8 = 5
    INT32 = 6
    UINT32 = 7
    FP8E5 = 8


class WeightGroupMode(Enum):
    """Dequantization weight group modes."""

    NONE = 0
    SHIFT = 1
    SYMMETRIC_NO_SHIFT = 2
    SYMMETRIC_WITH_SHIFT = 3
    ASYMMETRIC = 4


# Triton can only access global values instantiated as tl.constexpr
_WEIGHT_GROUP_MODE_NONE: tl.constexpr = tl.constexpr(WeightGroupMode.NONE.value)
_WEIGHT_GROUP_MODE_SHIFT: tl.constexpr = tl.constexpr(WeightGroupMode.SHIFT.value)
_WEIGHT_GROUP_MODE_SYMMETRIC_NO_SHIFT: tl.constexpr = tl.constexpr(WeightGroupMode.SYMMETRIC_NO_SHIFT.value)
_WEIGHT_GROUP_MODE_SYMMETRIC_WITH_SHIFT: tl.constexpr = tl.constexpr(WeightGroupMode.SYMMETRIC_WITH_SHIFT.value)
_WEIGHT_GROUP_MODE_ASYMMETRIC: tl.constexpr = tl.constexpr(WeightGroupMode.ASYMMETRIC.value)


class LoadOrder(Enum):
    """Order to load A matrix relative to other tensors."""

    VERY_EARLY = 0
    EARLY = 1
    MID = 2
    LATE = 3


# Triton can only access global values instantiated as tl.constexpr
_LOAD_ORDER_VERY_EARLY: tl.constexpr = tl.constexpr(LoadOrder.VERY_EARLY.value)
_LOAD_ORDER_EARLY: tl.constexpr = tl.constexpr(LoadOrder.EARLY.value)
_LOAD_ORDER_MID: tl.constexpr = tl.constexpr(LoadOrder.MID.value)
_LOAD_ORDER_LATE: tl.constexpr = tl.constexpr(LoadOrder.LATE.value)


class ChannelScaleMode(Enum):
    """Mode for channel scaling."""

    NONE = 0
    WEIGHT_ONLY = 1
    ACTIVATION_ONLY = 2
    WEIGHT_AND_ACTIVATION = 3


# Triton can only access global values instantiated as tl.constexpr
_CHANNEL_SCALE_MODE_NONE: tl.constexpr = tl.constexpr(ChannelScaleMode.NONE.value)
_CHANNEL_SCALE_MODE_WEIGHT_ONLY: tl.constexpr = tl.constexpr(ChannelScaleMode.WEIGHT_ONLY.value)
_CHANNEL_SCALE_MODE_ACTIVATION_ONLY: tl.constexpr = tl.constexpr(ChannelScaleMode.ACTIVATION_ONLY.value)
_CHANNEL_SCALE_MODE_WEIGHT_AND_ACTIVATION: tl.constexpr = tl.constexpr(ChannelScaleMode.WEIGHT_AND_ACTIVATION.value)


DTYPE_TO_TORCH = {
    0: torch.float32,
    1: torch.float16,
    2: torch.bfloat16,
    3: torch.float8_e4m3fn,
    4: torch.int8,
    5: torch.uint8,
    6: torch.int32,
    7: torch.uint32,
    8: torch.float8_e5m2,
    9: torch.float8_e4m3fnuz,
}

TORCH_DTYPE_TO_TRITON = {
    torch.float16: tl.float16,
    torch.float32: tl.float32,
    torch.bfloat16: tl.bfloat16,
    torch.int8: tl.int8,
    torch.uint8: tl.uint8,
    torch.int16: tl.int16,
    torch.uint16: tl.uint16,
    torch.int32: tl.int32,
    torch.uint32: tl.uint32,
    torch.float8_e4m3fn: tl.float8e4nv,
    torch.float8_e5m2: tl.float8e5,
    torch.float8_e4m3fnuz: tl.float8e4b8,
}

DTYPE_TO_TRITON = {key: TORCH_DTYPE_TO_TRITON[dtype] for key, dtype in DTYPE_TO_TORCH.items()}


def _get_matrix_a_eviction_policy() -> str:
    """Get eviction policy for Matrix A based on platform."""
    if current_platform.is_nvidia():
        return "evict_last"

    return ""


def _get_matrix_b_eviction_policy() -> str:
    """Get eviction policy for Matrix B based on platform."""
    if current_platform.is_nvidia():
        return "evict_first"

    return ""


def _get_metadata_eviction_policy() -> str:
    """Get eviction policy for metadata (scales and zeros) based on platform."""
    return ""


def _get_tuning_parameters() -> dict[str, int]:
    """Get block sizes/tuning parameters for current device."""
    device_name = current_platform.get_device_name()

    if "H100" in device_name:
        return {
            "cxpr_block_size_m": 128,
            "cxpr_block_size_n": 128,
            "cxpr_block_size_k": 128,
            "cxpr_group_size_m": 8,
            "num_warps": 8,
            "num_stages": 2,
        }

    if "MI300X" in device_name:
        return {
            "cxpr_block_size_m": 128,
            "cxpr_block_size_n": 64,
            "cxpr_block_size_k": 128,
            "cxpr_group_size_m": 16,
            "num_warps": 8,
            "num_stages": 2,
        }

    return {
        "cxpr_block_size_m": 64,
        "cxpr_block_size_n": 64,
        "cxpr_block_size_k": 32,
        "cxpr_group_size_m": 8,
    }


@triton.jit  # type: ignore[misc]
def _swizzle_tile(
    pid: int,
    m_dim: int,
    n_dim: int,
    cxpr_block_size_m: tl.constexpr,
    cxpr_block_size_n: tl.constexpr,
    cxpr_group_size_m: tl.constexpr,
) -> tuple[int, int]:
    """Return pid based on a swizzled tile of size (m, n)."""
    grid_m = tl.cdiv(m_dim, cxpr_block_size_m)
    grid_n = tl.cdiv(n_dim, cxpr_block_size_n)
    width = cxpr_group_size_m * grid_n
    group_id = pid // width
    group_size = tl.minimum(grid_m - group_id * cxpr_group_size_m, cxpr_group_size_m)
    pid_m = group_id * cxpr_group_size_m + (pid % group_size)
    pid_n = (pid % width) // group_size
    return pid_m, pid_n


@triton.jit  # type: ignore[misc]
def _linear_tile(pid: int, n_dim: int, cxpr_block_size_n: tl.constexpr) -> tuple[int, int]:
    """Return pid based on a linear tile of size (m, n)."""
    pid_m = pid // tl.cdiv(n_dim, cxpr_block_size_n)
    pid_n = pid % tl.cdiv(n_dim, cxpr_block_size_n)
    return pid_m, pid_n


@triton.jit  # type: ignore[misc]
def _dequantize(
    b: tl.tensor,
    scales: tl.tensor,
    zeros: tl.tensor,
    q_shift: tl.tensor,
    cxpr_input_dtype: tl.constexpr,
    cxpr_meta_dtype: tl.constexpr,
    cxpr_unpack_mask: tl.constexpr,
    cxpr_elements_per_sample: tl.constexpr,
    cxpr_w_group_mode: tl.constexpr,
    cxpr_zero_is_scalar: tl.constexpr,
    cxpr_weight_bias: tl.constexpr,
) -> tl.tensor:
    """Dequantize a tensor."""
    # Unpack
    if cxpr_elements_per_sample > 1:
        b = ((b >> q_shift) & cxpr_unpack_mask).to(cxpr_meta_dtype)
        b -= cxpr_weight_bias

    # Shift (operation: b - zeros)
    if cxpr_w_group_mode == _WEIGHT_GROUP_MODE_SHIFT:
        b -= zeros

    # Symmetric no shift (operation: b * scales)
    if cxpr_w_group_mode == _WEIGHT_GROUP_MODE_SYMMETRIC_NO_SHIFT:
        b = b.to(cxpr_meta_dtype) * scales

    # Asymmetric / Symmetric with shift (operation: (b - zeros) * scales)
    if cxpr_w_group_mode == _WEIGHT_GROUP_MODE_SYMMETRIC_WITH_SHIFT:
        b = (
            (b - zeros).to(cxpr_meta_dtype) * scales
            if cxpr_zero_is_scalar
            else (b.to(cxpr_meta_dtype) - zeros) * scales
        )

    # Asymmetric (operation: b * scales + zeros)
    if cxpr_w_group_mode == _WEIGHT_GROUP_MODE_ASYMMETRIC:
        b = tl.fma(b.to(cxpr_meta_dtype), scales, zeros)

    return b.to(cxpr_input_dtype)


@triton.jit  # type: ignore[misc]
def _gemm_kernel(
    a_ptr: tl.tensor,
    b_ptr: tl.tensor,
    c_ptr: tl.tensor,
    scales_ptr: tl.tensor,
    zeros_ptr: tl.tensor,
    scales_a_ptr: tl.tensor,
    # Input sizes
    m_dim: int,
    k_dim: int,
    n_dim: int,
    # Quantization parameters
    cxpr_w_nbits: tl.constexpr,
    cxpr_weight_bias: tl.constexpr,
    cxpr_group_size: tl.constexpr,
    cxpr_unpack_mask: tl.constexpr,
    cxpr_elements_per_sample: tl.constexpr,
    # Tensor strides
    matrix_a_stride_m: int,
    matrix_a_stride_k: int,
    matrix_b_stride_k: int,
    matrix_b_stride_n: int,
    matrix_c_stride_m: int,
    matrix_c_stride_n: int,
    meta_stride_g: int,
    meta_stride_n: int,
    # Data types
    cxpr_input_dtype: tl.constexpr,
    cxpr_output_dtype: tl.constexpr,
    cxpr_acc_dtype: tl.constexpr,
    cxpr_meta_dtype: tl.constexpr,
    # Metadata modes
    cxpr_channel_scale_mode: tl.constexpr,
    cxpr_w_group_mode: tl.constexpr,
    cxpr_zero_is_scalar: tl.constexpr,
    # Tuning parameters
    cxpr_block_size_m: tl.constexpr,
    cxpr_block_size_n: tl.constexpr,
    cxpr_block_size_k: tl.constexpr,
    cxpr_group_size_m: tl.constexpr,
    cxpr_block_size_scale_a: tl.constexpr,
    cxpr_block_size_scale_b: tl.constexpr,
    cxpr_matrix_a_load_order: tl.constexpr,
    cxpr_matrix_a_eviction_policy: tl.constexpr,
    cxpr_matrix_b_eviction_policy: tl.constexpr,
    cxpr_meta_eviction_policy: tl.constexpr,
    cxpr_data_contiguous: tl.constexpr,
    cxpr_swizzle_pid: tl.constexpr,
) -> None:
    """Based on https://github.com/fpgaminer/GPTQ-triton.

    GEMM for C = matmul(A, dequantize(B, scales, zeros))
    A is of shape (m_dim, k_dim): float16 or bfloat16
    B is of shape (k_dim//elements_per_sample, n_dim): int32 as a packed matrix
    C is of shape (m_dim, n_dim): float16 or bfloat16 depending on the input A
    scales and zeros is of shape (k_dim // group_size, n_dim): float16 or bfloat16
    cxpr_block_size_m >=16
    cxpr_block_size_k <= group_size
    """
    pid = tl.program_id(axis=0)

    pid_m, pid_n = (
        _swizzle_tile(pid, m_dim, n_dim, cxpr_block_size_m, cxpr_block_size_n, cxpr_group_size_m)
        if cxpr_swizzle_pid
        else _linear_tile(pid, n_dim, cxpr_block_size_n)
    )

    num_groups = tl.cdiv(k_dim, cxpr_block_size_k)

    matrix_a_offsets_m = pid_m * cxpr_block_size_m + tl.arange(0, cxpr_block_size_m)
    scale_a_offsets_m = (
        pid_m * cxpr_block_size_scale_a + tl.arange(0, cxpr_block_size_scale_a)
        if cxpr_block_size_scale_a > 1
        else tl.arange(0, 1)
    )

    matrix_b_offsets_n = pid_n * cxpr_block_size_n + tl.arange(0, cxpr_block_size_n)
    scale_b_offsets_n = (
        pid_n * cxpr_block_size_scale_b + tl.arange(0, cxpr_block_size_scale_b)
        if cxpr_block_size_scale_b > 1
        else tl.arange(0, 1)
    )

    group_offsets = tl.arange(0, cxpr_block_size_k)

    # Vectorized coalesced load
    if cxpr_data_contiguous:
        matrix_b_offsets_n = tl.max_contiguous(tl.multiple_of(matrix_b_offsets_n, cxpr_block_size_n), cxpr_block_size_n)

    # Calculate offsets for pointer into the block of matrix A
    matrix_a_block_offsets = (
        matrix_a_offsets_m[:, None] * matrix_a_stride_m + group_offsets[None, :] * matrix_a_stride_k
    )
    # Create pointer to block of matrix A
    matrix_a_block_ptr = a_ptr + matrix_a_block_offsets
    # Mask out any out-of-bounds elements
    matrix_a_block_mask = matrix_a_offsets_m[:, None] < m_dim

    # Calculate offsets for pointer into the block of matrix B
    matrix_b_block_offsets = (
        group_offsets[:, None] // cxpr_elements_per_sample
    ) * matrix_b_stride_k + matrix_b_offsets_n[None, :] * matrix_b_stride_n
    # Create pointer to block of matrix B
    matrix_b_block_ptr = b_ptr + matrix_b_block_offsets

    # Quantization metadata
    q_shift = ((group_offsets % cxpr_elements_per_sample) * cxpr_w_nbits).to(tl.int32)[:, None]
    scales_block_ptr = scales_ptr + matrix_b_offsets_n[None, :] * meta_stride_n
    zeros_block_ptr = zeros_ptr + matrix_b_offsets_n[None, :] * meta_stride_n
    stride_mul = cxpr_block_size_k / cxpr_group_size

    if cxpr_zero_is_scalar:
        zero_scalar = tl.load(zeros_ptr, eviction_policy=cxpr_matrix_a_eviction_policy)

    # Output accumulator
    acc = tl.zeros((cxpr_block_size_m, cxpr_block_size_n), dtype=cxpr_acc_dtype)

    for group_index in range(num_groups):
        # Very early load
        if cxpr_matrix_a_load_order == _LOAD_ORDER_VERY_EARLY:
            a = tl.load(
                matrix_a_block_ptr, mask=matrix_a_block_mask, other=0.0, eviction_policy=cxpr_matrix_a_eviction_policy
            )

        b = tl.load(matrix_b_block_ptr, eviction_policy=cxpr_matrix_b_eviction_policy)

        # Early load
        if cxpr_matrix_a_load_order == _LOAD_ORDER_EARLY:
            a = tl.load(
                matrix_a_block_ptr, mask=matrix_a_block_mask, other=0.0, eviction_policy=cxpr_matrix_a_eviction_policy
            )

        # Load meta-data
        if cxpr_w_group_mode > _WEIGHT_GROUP_MODE_NONE:
            group_offset = (group_index * stride_mul).to(tl.int32) * meta_stride_g

        # Load scales if they exist
        scales = None
        if cxpr_w_group_mode >= _WEIGHT_GROUP_MODE_SYMMETRIC_NO_SHIFT:
            scales = tl.load(scales_block_ptr + group_offset, eviction_policy=cxpr_meta_eviction_policy)

        # Load zeros if they exist
        zeros = None
        if (
            cxpr_w_group_mode == _WEIGHT_GROUP_MODE_SHIFT
            or cxpr_w_group_mode >= _WEIGHT_GROUP_MODE_SYMMETRIC_WITH_SHIFT
        ):
            zeros = (
                zero_scalar
                if cxpr_zero_is_scalar
                else tl.load(zeros_block_ptr + group_offset, eviction_policy=cxpr_meta_eviction_policy)
            ).to(cxpr_meta_dtype)

        # Mid load
        if cxpr_matrix_a_load_order == _LOAD_ORDER_MID:
            a = tl.load(
                matrix_a_block_ptr, mask=matrix_a_block_mask, other=0.0, eviction_policy=cxpr_matrix_a_eviction_policy
            )

        # Unpack and dequantize
        b = _dequantize(
            b,
            scales,
            zeros,
            q_shift,
            cxpr_input_dtype,
            cxpr_meta_dtype,
            cxpr_unpack_mask,
            cxpr_elements_per_sample,
            cxpr_w_group_mode,
            cxpr_zero_is_scalar,
            cxpr_weight_bias,
        )

        # Late load
        if cxpr_matrix_a_load_order == _LOAD_ORDER_LATE:
            a = tl.load(
                matrix_a_block_ptr, mask=matrix_a_block_mask, other=0.0, eviction_policy=cxpr_matrix_a_eviction_policy
            )

        # Dot
        acc = tl.dot(a, b, acc=acc, out_dtype=cxpr_acc_dtype)

        # Advance
        matrix_a_block_ptr += cxpr_block_size_k * matrix_a_stride_k
        matrix_b_block_ptr += (cxpr_block_size_k // cxpr_elements_per_sample) * matrix_b_stride_k

    # Channel-wise scaling
    if cxpr_channel_scale_mode == _CHANNEL_SCALE_MODE_WEIGHT_ONLY:
        scales_b = tl.load(
            scales_ptr + matrix_b_offsets_n,
            mask=matrix_b_offsets_n < n_dim,
            other=1,
            eviction_policy=cxpr_meta_eviction_policy,
        )
        acc = acc.to(cxpr_meta_dtype) * scales_b[None, :]

    if cxpr_channel_scale_mode == _CHANNEL_SCALE_MODE_ACTIVATION_ONLY:
        scales_a = tl.load(
            scales_a_ptr + matrix_a_offsets_m,
            mask=matrix_a_offsets_m < m_dim,
            other=1,
            eviction_policy=cxpr_meta_eviction_policy,
        )
        scales_b = tl.full((cxpr_block_size_n,), value=1, dtype=cxpr_meta_dtype)
        acc = acc.to(cxpr_meta_dtype) * (scales_a[:, None] * scales_b[None, :])

    if cxpr_channel_scale_mode == _CHANNEL_SCALE_MODE_WEIGHT_AND_ACTIVATION:
        scales_a = tl.load(
            scales_a_ptr + scale_a_offsets_m,
            mask=scale_a_offsets_m < m_dim,
            other=1,
            eviction_policy=cxpr_meta_eviction_policy,
        )
        scales_b = tl.load(
            scales_ptr + scale_b_offsets_n,
            mask=scale_b_offsets_n < n_dim,
            other=1,
            eviction_policy=cxpr_meta_eviction_policy,
        )
        acc = acc.to(cxpr_meta_dtype) * (scales_a[:, None] * scales_b[None, :])

    # Calculate offsets for pointer into the block of the output matrix
    matrix_c_block_offsets_m = pid_m * cxpr_block_size_m + tl.arange(0, cxpr_block_size_m)
    matrix_c_block_offsets_n = pid_n * cxpr_block_size_n + tl.arange(0, cxpr_block_size_n)
    matrix_c_block_offsets_n = tl.max_contiguous(
        tl.multiple_of(matrix_c_block_offsets_n, cxpr_block_size_n), cxpr_block_size_n
    )

    # Create pointer to block of output matrix to store accumulated result
    matrix_c_block_ptr = c_ptr + (
        matrix_c_block_offsets_m[:, None] * matrix_c_stride_m + matrix_c_block_offsets_n[None, :] * matrix_c_stride_n
    )
    # Create mask of elements to store
    matrix_c_block_mask = (matrix_c_block_offsets_m[:, None] < m_dim) & (matrix_c_block_offsets_n[None, :] < n_dim)

    # Store accumulated result
    tl.store(matrix_c_block_ptr, acc.to(cxpr_output_dtype), mask=matrix_c_block_mask)


@dataclass
class MixedPrecisionMatmulMetadata:
    """Metadata for GEMM kernel."""

    m_dim: int
    k_dim: int
    n_dim: int
    weight_size_bits: int
    weight_bias: int
    group_size: int
    elements_per_sample: int
    zero_is_scalar: bool
    unpack_mask: int
    data_contiguous: bool
    input_dtype: torch.dtype
    output_dtype: torch.dtype
    acc_dtype: torch.dtype
    meta_dtype: torch.dtype
    channel_scale_mode: ChannelScaleMode
    weight_group_mode: WeightGroupMode


def mixed_precision_gemm_launcher(
    output: torch.Tensor,
    x: torch.Tensor,
    w_q_packed: torch.Tensor,
    scales: torch.Tensor,
    zeros: torch.Tensor | None,
    metadata: MixedPrecisionMatmulMetadata,
) -> None:
    """GEMM Triton Launcher."""
    grid = lambda parameters: (  # noqa: E731
        triton.cdiv(metadata.m_dim, parameters["cxpr_block_size_m"])
        * triton.cdiv(metadata.n_dim, parameters["cxpr_block_size_n"]),
    )

    zeros = zeros if zeros is not None else torch.tensor([[]], dtype=torch.int32, device=x.device)

    tuning_parameters = _get_tuning_parameters()

    _gemm_kernel[grid](
        a_ptr=x,
        b_ptr=w_q_packed,
        c_ptr=output,
        scales_ptr=scales,
        zeros_ptr=zeros,
        scales_a_ptr=None,
        # Tensor sizes
        m_dim=metadata.m_dim,
        k_dim=metadata.k_dim,
        n_dim=metadata.n_dim,
        # Quantization paramers
        cxpr_w_nbits=metadata.weight_size_bits,
        cxpr_weight_bias=metadata.weight_bias,
        cxpr_group_size=metadata.group_size,
        cxpr_unpack_mask=metadata.unpack_mask,
        cxpr_elements_per_sample=metadata.elements_per_sample,
        # Strides
        matrix_a_stride_m=x.stride(0),
        matrix_a_stride_k=x.stride(1),
        matrix_b_stride_k=w_q_packed.stride(0),
        matrix_b_stride_n=w_q_packed.stride(1),
        matrix_c_stride_m=output.stride(0),
        matrix_c_stride_n=output.stride(1),
        meta_stride_g=scales.stride(0),
        meta_stride_n=scales.stride(1),
        # Data types
        cxpr_input_dtype=TORCH_DTYPE_TO_TRITON[metadata.input_dtype],
        cxpr_output_dtype=TORCH_DTYPE_TO_TRITON[metadata.output_dtype],
        cxpr_acc_dtype=TORCH_DTYPE_TO_TRITON[metadata.acc_dtype],
        cxpr_meta_dtype=TORCH_DTYPE_TO_TRITON[metadata.meta_dtype],
        # Metadata modes
        cxpr_channel_scale_mode=metadata.channel_scale_mode.value,
        cxpr_w_group_mode=metadata.weight_group_mode.value,
        cxpr_zero_is_scalar=metadata.zero_is_scalar,
        # Tuning parameters
        cxpr_data_contiguous=metadata.data_contiguous,
        cxpr_matrix_a_load_order=LoadOrder.MID.value,
        cxpr_matrix_a_eviction_policy=_get_matrix_a_eviction_policy(),
        cxpr_matrix_b_eviction_policy=_get_matrix_b_eviction_policy(),
        cxpr_meta_eviction_policy=_get_metadata_eviction_policy(),
        cxpr_swizzle_pid=True,
        cxpr_block_size_scale_a=tuning_parameters["cxpr_block_size_m"],
        cxpr_block_size_scale_b=tuning_parameters["cxpr_block_size_n"],
        **tuning_parameters,
    )


@dataclass
class ScaledMatmulMetadata:
    """Metadata for GEMM kernel."""

    m_dim: int
    k_dim: int
    n_dim: int
    data_contiguous: bool
    input_dtype: torch.dtype
    output_dtype: torch.dtype
    acc_dtype: torch.dtype
    meta_dtype: torch.dtype
    channel_scale_mode: ChannelScaleMode
    weight_group_mode: WeightGroupMode


def scaled_gemm_launcher(
    output: torch.Tensor,
    a: torch.Tensor,
    b: torch.Tensor,
    scale_a: torch.Tensor,
    scale_b: torch.Tensor,
    metadata: ScaledMatmulMetadata,
) -> None:
    """GEMM Triton Launcher."""
    grid = lambda parameters: (  # noqa: E731
        triton.cdiv(metadata.m_dim, parameters["cxpr_block_size_m"])
        * triton.cdiv(metadata.n_dim, parameters["cxpr_block_size_n"]),
    )

    zeros = torch.tensor([[]], dtype=torch.int32, device=a.device)

    tuning_parameters = _get_tuning_parameters()

    _gemm_kernel[grid](
        a_ptr=a,
        b_ptr=b,
        c_ptr=output,
        scales_ptr=scale_b,
        zeros_ptr=zeros,
        scales_a_ptr=scale_a,
        # Tensor sizes
        m_dim=metadata.m_dim,
        k_dim=metadata.k_dim,
        n_dim=metadata.n_dim,
        # Quantization paramers
        cxpr_w_nbits=0,
        cxpr_weight_bias=0,
        cxpr_group_size=1,
        cxpr_unpack_mask=0,
        cxpr_elements_per_sample=1,
        cxpr_zero_is_scalar=True,
        # Strides
        matrix_a_stride_m=a.stride(0),
        matrix_a_stride_k=a.stride(1),
        matrix_b_stride_k=b.stride(0),
        matrix_b_stride_n=b.stride(1),
        matrix_c_stride_m=output.stride(0),
        matrix_c_stride_n=output.stride(1),
        meta_stride_g=1 if scale_b.numel() == 1 else scale_b.stride(0),
        meta_stride_n=1 if scale_b.numel() == 1 else scale_b.stride(1),
        # Data types
        cxpr_input_dtype=TORCH_DTYPE_TO_TRITON[metadata.input_dtype],
        cxpr_output_dtype=TORCH_DTYPE_TO_TRITON[metadata.output_dtype],
        cxpr_acc_dtype=TORCH_DTYPE_TO_TRITON[metadata.acc_dtype],
        cxpr_meta_dtype=TORCH_DTYPE_TO_TRITON[metadata.meta_dtype],
        # Metadata modes
        cxpr_channel_scale_mode=metadata.channel_scale_mode.value,
        cxpr_w_group_mode=metadata.weight_group_mode.value,
        # Tuning parameters
        cxpr_data_contiguous=metadata.data_contiguous,
        cxpr_matrix_a_load_order=LoadOrder.MID.value,
        cxpr_matrix_a_eviction_policy=_get_matrix_a_eviction_policy(),
        cxpr_matrix_b_eviction_policy=_get_matrix_b_eviction_policy(),
        cxpr_meta_eviction_policy=_get_metadata_eviction_policy(),
        cxpr_swizzle_pid=True,
        cxpr_block_size_scale_a=1 if scale_a.numel() == 1 else tuning_parameters["cxpr_block_size_m"],
        cxpr_block_size_scale_b=1 if scale_b.numel() == 1 else tuning_parameters["cxpr_block_size_n"],
        **tuning_parameters,
    )



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Reference implementation of RMS norm kernel."""

import torch

from conch import envs
from conch.platforms import current_platform


def _rms_norm_pytorch_ref(x: torch.Tensor, weight: torch.Tensor, epsilon: float) -> torch.Tensor:
    """PyTorch reference rms_norm impl."""
    orig_dtype = x.dtype
    x = x.to(torch.float32)
    variance = x.pow(2).mean(dim=-1, keepdim=True)
    x = x * torch.rsqrt(variance + epsilon)
    # Llama does x.to(float16) * w whilst Gemma is (x * w).to(float16)
    # See https://github.com/huggingface/transformers/pull/29402
    x = x.to(orig_dtype) * weight
    return x


def _rms_norm_vllm_ref(x: torch.Tensor, weight: torch.Tensor, epsilon: float) -> torch.Tensor:
    """vLLM reference rms_norm impl."""
    from vllm._custom_ops import rms_norm as rms_norm_cuda

    out = torch.empty_like(x, dtype=x.dtype, device=x.device)
    rms_norm_cuda(out, x, weight, epsilon)

    return out


def _fused_add_rms_norm_pytorch_ref(
    x: torch.Tensor,
    residual: torch.Tensor,
    weight: torch.Tensor,
    epsilon: float,
) -> tuple[torch.Tensor, torch.Tensor]:
    """PyTorch reference fused_add_rms_norm impl."""
    x = x + residual
    residual = x

    orig_dtype = x.dtype
    x = x.to(torch.float32)
    variance = x.pow(2).mean(dim=-1, keepdim=True)
    x = x * torch.rsqrt(variance + epsilon)
    # Llama does x.to(float16) * w whilst Gemma is (x * w).to(float16)
    # See https://github.com/huggingface/transformers/pull/29402
    x = x.to(orig_dtype) * weight
    return x, residual


def _fused_add_rms_norm_vllm_ref(
    x: torch.Tensor,
    residual: torch.Tensor,
    weight: torch.Tensor,
    epsilon: float,
) -> tuple[torch.Tensor, torch.Tensor]:
    """vLLM reference fused_add_rms_norm impl."""
    from vllm._custom_ops import fused_add_rms_norm as fused_add_rms_norm_cuda

    fused_add_rms_norm_cuda(x, residual, weight, epsilon)

    return x, residual  # type: ignore[no-any-return, unused-ignore]


def rms_norm(x: torch.Tensor, weight: torch.Tensor, epsilon: float) -> torch.Tensor:
    """RMS norm operation."""
    if envs.CONCH_ENABLE_VLLM and current_platform.has_cuda():
        return _rms_norm_vllm_ref(
            x,
            weight,
            epsilon,
        )

    return _rms_norm_pytorch_ref(
        x,
        weight,
        epsilon,
    )


def fused_add_rms_norm(
    x: torch.Tensor,
    residual: torch.Tensor,
    weight: torch.Tensor,
    epsilon: float,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Fused add RMS norm operation."""
    if envs.CONCH_ENABLE_VLLM and current_platform.has_cuda():
        return _fused_add_rms_norm_vllm_ref(
            x,
            residual,
            weight,
            epsilon,
        )

    return _fused_add_rms_norm_pytorch_ref(
        x,
        residual,
        weight,
        epsilon,
    )



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

from typing import Final

from setuptools import setup

_TORCH_VERSION: Final = "2.7.0"

_REQUIREMENTS: Final = [
    "numpy>=1.26.4",
]

# For CPU:
# --extra-index-url https://download.pytorch.org/whl/cpu
_DEFAULT_PLATFORM_REQUIREMENTS: Final = [
    f"torch=={_TORCH_VERSION}",
    "triton>=3.3.0",
]

# For ROCm:
# --extra-index-url https://download.pytorch.org/whl/rocm6.2.4
_ROCM_PLATFORM_REQUIREMENTS: Final = [
    f"torch=={_TORCH_VERSION}+rocm6.2.4",
    "pytorch-triton-rocm>=3.3.0",
]

# For XPU:
# --extra-index-url https://download.pytorch.org/whl/xpu
_XPU_PLATFORM_REQUIREMENTS: Final = [
    f"torch=={_TORCH_VERSION}+xpu",
    "pytorch-triton-xpu>=3.3.0",
]

_PLATFORM_REQUIREMENTS: Final = {
    # --extra-index-url https://download.pytorch.org/whl/cpu
    "cpu": _DEFAULT_PLATFORM_REQUIREMENTS,
    # No extra index URL needed!
    "cuda": _DEFAULT_PLATFORM_REQUIREMENTS,
    # --extra-index-url https://download.pytorch.org/whl/rocm6.2.4
    "rocm": _ROCM_PLATFORM_REQUIREMENTS,
    # --extra-index-url https://download.pytorch.org/whl/xpu
    "xpu": _XPU_PLATFORM_REQUIREMENTS,
}


def get_default_dependencies() -> list[str]:
    """Determine the appropriate dependencies based on detected hardware."""
    return _REQUIREMENTS


def get_optional_dependencies() -> dict[str, list[str]]:
    """Get optional dependency groups."""
    return {
        "dev": [
            "click>=8.1.8",
            "coverage>=7.8.0",
            "einops>=0.8.0",
            "matplotlib>=3.10.1",
            "mypy>=1.15.0",
            "pandas>=2.2.3",
            "pre-commit>=4.2.0",
            "pytest>=8.3.4",
            "ruff>=0.4.10",
        ],
    } | _PLATFORM_REQUIREMENTS


setup(  # type: ignore[no-untyped-call]
    name="conch-triton-kernels",
    install_requires=get_default_dependencies(),
    extras_require=get_optional_dependencies(),
    setup_requires=["wheel"],
)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Reference implementation of silu and mul kernel."""

import torch
import torch.nn.functional as F  # noqa: N812

from conch import envs
from conch.platforms import current_platform


def _silu_and_mul_pytorch_ref(x: torch.Tensor) -> torch.Tensor:
    """PyTorch reference silu and mul implementation."""
    d = x.shape[-1] // 2
    return F.silu(x[..., :d]) * x[..., d:]


def _silu_and_mul_vllm_ref(x: torch.Tensor) -> torch.Tensor:
    """vLLM reference silu and mul implementation."""
    from vllm.model_executor.layers.activation import SiluAndMul

    silu_layer = SiluAndMul()  # type: ignore[no-untyped-call, unused-ignore]
    return silu_layer.forward_cuda(x)  # type: ignore[no-any-return, unused-ignore]


def silu_and_mul(x: torch.Tensor) -> torch.Tensor:
    """Silu and mul operation."""
    if envs.CONCH_ENABLE_VLLM and current_platform.has_cuda():
        return _silu_and_mul_vllm_ref(x)

    return _silu_and_mul_pytorch_ref(x)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Conch rms_norm benchmark."""

import sys
from typing import Final

import click
import torch

from conch.ops.normalization.rms_norm import rms_norm as rms_norm_conch
from conch.platforms import current_platform
from conch.reference.normalization.rms_norm import rms_norm as rms_norm_reference
from conch.third_party.vllm.utils import seed_everything
from conch.utils.benchmark import BenchmarkMetadata, benchmark_it


@click.command()
@click.option(
    "--hidden-size",
    required=True,
    type=int,
    default=2048,
    help="Dimension",
)
@click.option(
    "--num-tokens",
    required=True,
    type=int,
    default=4096,
    help="Number of tokens",
)
@click.option(
    "--iteration-time-ms",
    required=False,
    type=int,
    default=10000,
    help="Time in milliseconds to run benchmark",
)
@click.option(
    "--warmup-time-ms",
    required=False,
    type=int,
    default=1000,
    help="Time in milliseconds to warmup before recording times",
)
@click.option(
    "--verbose",
    is_flag=True,
    help="Flag for printing verbose output",
)
@click.option(
    "--gpu",
    required=False,
    type=str,
    default=current_platform.device,
    help="Device to run on",
)
@click.option(
    "--csv",
    is_flag=True,
    help="Flag for printing results in CSV format",
)
def main(  # noqa: PLR0913
    hidden_size: int,
    num_tokens: int,
    iteration_time_ms: int,
    warmup_time_ms: int,
    verbose: bool,
    gpu: str,
    csv: bool,
) -> None:
    """Benchmark Conch rms_norm op.

    Args:
        hidden_size: Dimension of input/output.
        num_tokens: Number of tokens.
        iteration_time_ms: Time in milliseconds to run the benchmark.
        warmup_time_ms: Time in milliseconds to warm up before recording times.
        verbose: Flag to indicate whether or not to print verbose output.
        gpu: Which gpu to run on.
        csv: Flag for printing results in CSV format.
    """
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(gpu)
    torch.set_default_device(device)

    dtype: Final = torch.float16
    tolerance: Final = 1e-2
    epsilon: Final = 1e-6

    metadata = BenchmarkMetadata(
        platform=current_platform.name(),
        params={
            "hidden_size": hidden_size,
            "num_tokens": num_tokens,
        },
    )

    x = torch.randn((num_tokens, hidden_size), dtype=dtype, device=device)
    weight = torch.randn((hidden_size,), dtype=dtype, device=device)

    conch_output = rms_norm_conch(x, weight, epsilon)

    ref_output = rms_norm_reference(x, weight, epsilon)

    if not torch.allclose(ref_output, conch_output, atol=tolerance, rtol=tolerance):
        print(f"WARNING: Reference and Conch results differ! (atol={tolerance}, rtol={tolerance})", file=sys.stderr)
        print(f"Output max diff: {(ref_output - conch_output).abs().max().item()}", file=sys.stderr)

        if verbose:
            print(f"Reference output: {ref_output}", file=sys.stderr)
            print(f"Conch output: {conch_output}", file=sys.stderr)
    else:
        print(f"Results matched with atol={tolerance} and rtol={tolerance} :)", file=sys.stderr)

    # Benchmark Reference vs. Conch implementations
    baseline_result = benchmark_it(
        lambda: rms_norm_reference(
            x,
            weight,
            epsilon,
        ),
        tag="Baseline",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    conch_result = benchmark_it(
        lambda: rms_norm_conch(
            x,
            weight,
            epsilon,
        ),
        tag="Conch",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    # Print results
    conch_result.print_parameters(csv=csv)
    conch_result.print_results(csv=csv)
    baseline_result.print_results(csv=csv)


if __name__ == "__main__":
    main()



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Triton implementation of FP8 CompressedTensors scaling kernels."""

import torch
import triton
import triton.language as tl

from conch.platforms import current_platform


@triton.jit  # type: ignore[misc]
def _static_scaled_fp8_quant_kernel(
    # Pointers to tensors
    output_ptr: tl.tensor,  # (num_tokens, hidden_size)
    input_ptr: tl.tensor,  # (num_tokens, hidden_size)
    scale_ptr: tl.tensor,  # (1,)
    # Scalar arguments
    hidden_size: int,
    # Constexprs
    cxpr_hidden_size_padded: tl.constexpr,
    cxpr_block_size: tl.constexpr,
    cxpr_is_rocm: tl.constexpr,
) -> None:
    """FP8 quantization kernel using static scaling.

    Args:
        output_ptr: Pointer to tensor for output, shape: (num_tokens, hidden_size).
        input_ptr: Pointer to tensor for fp input, shape: (num_tokens, hidden_size).
        scale_ptr: Pointer to static scale factor, shape: (1,).
        hidden_size: Second dimension of input/output tensors.
        cxpr_hidden_size_padded: Hidden size padded to next power-of-two.
        cxpr_block_size: Block size to iterate through the hidden size for each token.
        cxpr_is_rocm: Whether or not we're on AMD.
    """
    # Program id determines which token we are processing
    token_idx = tl.program_id(0)
    # Calculate offset to this token from the start of the input/output tensors
    token_offset = token_idx * hidden_size

    # Common offsets that can be shared for each block
    block_offsets = tl.arange(0, cxpr_block_size)

    fp8_dtype = tl.float8e4b8 if cxpr_is_rocm else tl.float8e4nv

    # Invert scale factor so we can multiply instead of divide
    inverted_scale = 1.0 / tl.load(scale_ptr)

    # Iterate through the hidden_size for this token in chunks of size cxpr_block_size
    for hidden_start_idx in tl.static_range(0, cxpr_hidden_size_padded, cxpr_block_size):
        # Calculate absolute offsets and mask for this block
        hidden_offsets = hidden_start_idx + block_offsets
        mask = hidden_offsets < hidden_size
        offsets = token_offset + hidden_offsets

        # Load block from input tensor
        block = tl.load(input_ptr + offsets, mask=mask)
        # Apply inverted scaling factor and cast to FP8
        block = (block * inverted_scale).to(fp8_dtype)
        # Store result to output tensor
        tl.store(output_ptr + offsets, block, mask=mask)


def static_scaled_fp8_quant_launcher(
    output_tensor: torch.Tensor,
    input_tensor: torch.Tensor,
    scale: torch.Tensor,
) -> None:
    """Launch Triton kernel to perform static-scaled fp8 quantization.

    Args:
        output_tensor: Tensor to write the output of the scaling, shape: (num_tokens, hidden_size).
        input_tensor: Tensor with input to scale, shape: (num_tokens, hidden_size).
        scale: Tensor with static scaling factor to apply, shape: (1).
    """
    num_tokens, hidden_size = output_tensor.shape

    # Triton requires power-of-two sizes
    hidden_size_padded = triton.next_power_of_2(hidden_size)
    block_size: tl.constexpr = min(hidden_size_padded, 1024)

    is_rocm: tl.constexpr = current_platform.is_amd()

    # Parallelize over the number of tokens in the sequence
    grid = (num_tokens,)

    # Launch kernel
    _static_scaled_fp8_quant_kernel[grid](
        output_ptr=output_tensor,
        input_ptr=input_tensor,
        scale_ptr=scale,
        hidden_size=hidden_size,
        cxpr_hidden_size_padded=hidden_size_padded,
        cxpr_block_size=block_size,
        cxpr_is_rocm=is_rocm,
    )



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Port of vllm gemma rms norm to Triton."""

import torch
import triton
import triton.language as tl


@triton.jit  # type: ignore[misc]
def _gemma_rms_norm_inplace_kernel(
    x_ptr: tl.tensor, weights_ptr: tl.const, hidden_size: int, eps: float, cxpr_block_size: tl.constexpr
) -> None:
    """Perform Gemma's version of RMS layer norm.

    Args:
        x_ptr: Tensor to be normalized, shape: (num_tokens, hidden_size).
        weights_ptr: learned weights of norm layer, shape: (hidden_size,).
        hidden_size: hidden size, often corresponds to head size.
        eps: value to pad during inverse-rms calculation.
        cxpr_block_size: must be next-power-of-two >= hidden_size.
    """
    # One block per row/token
    token_id = tl.program_id(0)
    token_row_ptr = x_ptr + token_id * hidden_size

    # Gemma RMS details
    # - Compute in float32, then convert back to original type
    # - mean-of-squares = mean(x ** 2)
    # - root-of-mean-of-squares (RMS) = sqrt(mean-of-squares + eps)
    # - weigh scaled inputs: x = (x / RMS) * (1 + weight)
    # fp32 as working format is discussed here:
    # https://github.com/huggingface/transformers/pull/29402
    # It would be reasonable do something a little bit more precise,
    # i.e. basing the precision on the input precision.  On the
    # other hand, fp32 is probably the most efficient type on most
    # GPUs' compute cores.

    offsets = tl.arange(0, cxpr_block_size)
    mask = offsets < hidden_size
    x = tl.load(token_row_ptr + offsets, mask=mask).to(tl.float32)
    x_sq = x * x
    mean_squares = tl.sum(x_sq) / hidden_size
    recip_rms = tl.rsqrt(mean_squares + eps)
    w = tl.load(weights_ptr + offsets, mask=mask).to(tl.float32)
    x = x * recip_rms * (1.0 + w)
    tl.store(token_row_ptr + offsets, x, mask=mask)  # Implicit cast back to original dtype


def gemma_rms_norm_inplace_launcher(
    x: torch.Tensor,
    weights: torch.Tensor,
    epsilon: float = 1e-6,
) -> None:
    """Perform Gemma's version of RMS layer norm.

    Args:
        x: Tensor to be normalized, shape: (num_tokens, hidden_size).
        weights: learned weights of norm layer, shape: (hidden_size,).
        epsilon: value to pad during inverse-rms calculation.
    """
    # Sanity check that the hidden dimensions match
    hidden_size = x.shape[-1]
    if hidden_size != weights.shape[-1]:
        msg = f"Input hidden dimenson ({hidden_size}) does not match length of weights ({weights.shape[-1]})"
        raise ValueError(msg)

    # Only support two-dimensional x
    if len(x.shape) != 2:  # noqa: PLR2004
        msg = f"x is {len(x.shape)}-dimensional.  Only supporting two dimensions."
        raise ValueError(msg)

    block_size = triton.next_power_of_2(hidden_size)
    grid = (x.shape[0],)
    _gemma_rms_norm_inplace_kernel[grid](
        x_ptr=x,
        weights_ptr=weights,
        hidden_size=hidden_size,
        eps=epsilon,
        cxpr_block_size=block_size,
    )



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Reference implementation of scaled GEMM."""

import torch

from conch import envs
from conch.platforms import current_platform


def _scaled_gemm_pytorch_ref(
    a: torch.Tensor,
    b: torch.Tensor,
    scale_a: torch.Tensor,
    scale_b: torch.Tensor,
    out_dtype: torch.dtype,
    bias: torch.Tensor | None = None,
) -> torch.Tensor:
    out = torch.matmul(a.to(torch.float32), b.to(torch.float32))
    out = scale_a * out
    out = scale_b.T * out
    out = out.to(out_dtype)
    if bias is not None:
        out = out + bias

    return out


def _scaled_gemm_vllm_ref(
    a: torch.Tensor,
    b: torch.Tensor,
    scale_a: torch.Tensor,
    scale_b: torch.Tensor,
    out_dtype: torch.dtype,
    bias: torch.Tensor | None = None,
) -> torch.Tensor:
    from vllm._custom_ops import cutlass_scaled_mm

    return cutlass_scaled_mm(a, b, scale_a, scale_b, out_dtype, bias)  # type: ignore[no-any-return, unused-ignore]


def scaled_gemm(
    a: torch.Tensor,
    b: torch.Tensor,
    scale_a: torch.Tensor,
    scale_b: torch.Tensor,
    out_dtype: torch.dtype,
    bias: torch.Tensor | None = None,
) -> torch.Tensor:
    # CUTLASS is only supported on Nvidia
    if envs.CONCH_ENABLE_VLLM and current_platform.is_nvidia():
        return _scaled_gemm_vllm_ref(
            a,
            b,
            scale_a,
            scale_b,
            out_dtype,
            bias,
        )

    return _scaled_gemm_pytorch_ref(
        a,
        b,
        scale_a,
        scale_b,
        out_dtype,
        bias,
    )



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, Any

from conch.platforms.platform import Platform, PlatformEnum, detect_current_platform

_current_platform = None

if TYPE_CHECKING:
    current_platform: Platform


def __getattr__(name: str) -> Any:
    if name == "current_platform":
        global _current_platform  # noqa: PLW0603
        if _current_platform is None:
            _current_platform = detect_current_platform()
        return _current_platform

    if name in globals():
        return globals()[name]

    error_msg = f"No attribute named '{name}' exists in {__name__}."
    raise AttributeError(error_msg)


__all__ = [
    "Platform",
    "PlatformEnum",
    "current_platform",
    "detect_current_platform",
]



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Reference implementation of gelu tanh and mul kernel."""

import torch
import torch.nn.functional as F  # noqa: N812

from conch import envs
from conch.platforms import current_platform


def _gelu_tanh_and_mul_pytorch_ref(x: torch.Tensor) -> torch.Tensor:
    """PyTorch reference gelu_tanh_and_mul impl."""
    d = x.shape[-1] // 2
    return F.gelu(x[..., :d], approximate="tanh") * x[..., d:]


def _gelu_tanh_and_mul_vllm_ref(x: torch.Tensor) -> torch.Tensor:
    """vLLM reference gelu_tanh_and_mul impl."""
    from vllm.model_executor.layers.activation import GeluAndMul

    gelu_layer = GeluAndMul("tanh")
    return gelu_layer.forward_cuda(x)  # type: ignore[no-any-return, unused-ignore]


def gelu_tanh_and_mul(x: torch.Tensor) -> torch.Tensor:
    """Gelu, tanh, and mul operation."""
    if envs.CONCH_ENABLE_VLLM and current_platform.has_cuda():
        return _gelu_tanh_and_mul_vllm_ref(x)

    return _gelu_tanh_and_mul_pytorch_ref(x)



# SPDX-License-Identifier: Apache-2.0

import random
from typing import Final

import torch

from conch import envs
from conch.platforms import current_platform

STR_DTYPE_TO_TORCH_DTYPE: Final = {
    "half": torch.half,
    "bfloat16": torch.bfloat16,
    "float": torch.float,
    "fp8": torch.uint8,
    "fp8_e4m3": torch.uint8,
    "fp8_e5m2": torch.uint8,
}


def seed_everything(seed: int) -> None:
    random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)


def _generate_random_fp8(
    tensor: torch.Tensor,
    low: float,
    high: float,
) -> None:
    # NOTE(zhaoyang): Due to NaN and Inf representation for fp8 data type,
    # it may occur Inf or NaN if we directly use torch.randint
    # to generate random data for fp8 data.
    # For example, s.11111.00 in fp8e5m2 format represents Inf.
    #     | E4M3        | E5M2
    # -----|-------------|-------------------
    # Inf | N/A         | s.11111.00
    # NaN | s.1111.111  | s.11111.{01,10,11}
    tensor_tmp = torch.empty_like(tensor, dtype=torch.float16)
    tensor_tmp.uniform_(low, high)
    if envs.CONCH_ENABLE_VLLM and current_platform.has_cuda():
        from vllm import _custom_ops as ops

        ops.convert_fp8(tensor, tensor_tmp)
    else:
        from conch.reference.quantization.fp8 import _scaled_fp8_quant_pytorch_ref

        tensor = _scaled_fp8_quant_pytorch_ref(tensor_tmp, scale=torch.tensor(1.0))
    del tensor_tmp


def get_kv_cache_torch_dtype(
    cache_dtype: str | torch.dtype | None, model_dtype: str | torch.dtype | None = None
) -> torch.dtype:
    if isinstance(cache_dtype, str):
        if cache_dtype == "auto":
            if isinstance(model_dtype, str):
                torch_dtype = STR_DTYPE_TO_TORCH_DTYPE[model_dtype]
            elif isinstance(model_dtype, torch.dtype):
                torch_dtype = model_dtype
            else:
                error_msg = f"Invalid model dtype: {model_dtype}"
                raise ValueError(error_msg)
        elif cache_dtype in ["half", "bfloat16", "float"]:
            torch_dtype = STR_DTYPE_TO_TORCH_DTYPE[cache_dtype]
        elif cache_dtype == "fp8":
            torch_dtype = torch.uint8
        else:
            error_msg = f"Invalid kv cache dtype: {cache_dtype}"
            raise ValueError(error_msg)
    elif isinstance(cache_dtype, torch.dtype):
        torch_dtype = cache_dtype
    else:
        error_msg = f"Invalid kv cache dtype: {cache_dtype}"
        raise TypeError(error_msg)
    return torch_dtype


def create_kv_caches_with_random(
    num_blocks: int,
    block_size: int,
    num_layers: int,
    num_heads: int,
    head_size: int,
    cache_dtype: str | torch.dtype | None,
    model_dtype: str | torch.dtype | None = None,
    seed: int = 0,
    device: str | torch.device = "cuda",
) -> tuple[list[torch.Tensor], list[torch.Tensor]]:
    if cache_dtype == "fp8" and head_size % 16:
        error_msg = f"Does not support key cache of type fp8 with head_size {head_size}"
        raise ValueError(error_msg)

    seed_everything(seed)

    torch_dtype = get_kv_cache_torch_dtype(cache_dtype, model_dtype)

    scale = head_size**-0.5
    x = 16 // torch.tensor([], dtype=torch_dtype).element_size()
    key_cache_shape = (num_blocks, num_heads, head_size // x, block_size, x)
    key_caches: list[torch.Tensor] = []
    for _ in range(num_layers):
        key_cache = torch.empty(size=key_cache_shape, dtype=torch_dtype, device=device)
        if cache_dtype in ["auto", "half", "bfloat16", "float"]:
            key_cache.uniform_(-scale, scale)
        elif cache_dtype == "fp8":
            _generate_random_fp8(key_cache, -scale, scale)
        else:
            error_msg = f"Does not support key cache of type {cache_dtype}"
            raise ValueError(error_msg)
        key_caches.append(key_cache)

    value_cache_shape = (num_blocks, num_heads, head_size, block_size)
    value_caches: list[torch.Tensor] = []
    for _ in range(num_layers):
        value_cache = torch.empty(size=value_cache_shape, dtype=torch_dtype, device=device)
        if cache_dtype in ["auto", "half", "bfloat16", "float"]:
            value_cache.uniform_(-scale, scale)
        elif cache_dtype == "fp8":
            _generate_random_fp8(value_cache, -scale, scale)
        else:
            error_msg = f"Does not support value cache of type {cache_dtype}"
            raise ValueError(error_msg)
        value_caches.append(value_cache)
    return key_caches, value_caches


def reshape_vllm_kvcache(
    key_cache_vllm: torch.Tensor, value_cache_vllm: torch.Tensor
) -> tuple[torch.Tensor, torch.Tensor]:
    """Reshape vLLM key and value caches into format expected by FlashAttention.

    Args:
        key_cache_vllm: vLLM key cache, shape: (num_cache_blocks, num_kv_heads, head_size // x, cache_block_size, x).
        value_cache_vllm: vLLM value cache, shape: (num_cache_blocks, num_kv_heads, head_size, cache_block_size).

    Returns:
        Reshaped key and value caches as (num_cache_blocks, cache_block_size, num_kv_heads, head_size).
    """
    num_cache_blocks, num_kv_heads, head_size, cache_block_size = value_cache_vllm.shape

    k = key_cache_vllm.permute(0, 1, 3, 2, 4).contiguous().reshape(num_cache_blocks, num_kv_heads, cache_block_size, head_size)
    v = value_cache_vllm.permute(0, 1, 3, 2).contiguous().reshape(num_cache_blocks, num_kv_heads, cache_block_size, head_size)

    k = k.permute(0, 2, 1, 3).contiguous()
    v = v.permute(0, 2, 1, 3).contiguous()

    return k, v


def create_tensors(
    head_dim: int,
    seq_len: int,
    cache_block_size: int,
    batch_size: int,
    num_query_heads: int,
    num_kv_heads: int,
    kv_cache_dtype: str,
    gpu: str,
    dtype: torch.dtype,
    seed: int = 0,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """Create Q, K, V, block tables, and sequence lengths tensors for benchmarking."""
    device: Final = torch.device(gpu)
    torch.set_default_device(device)

    assert num_query_heads % num_kv_heads == 0

    scale: Final = float(1.0 / (head_dim**0.5))
    total_num_cache_elements: Final = batch_size * seq_len
    num_cache_blocks: Final = total_num_cache_elements // cache_block_size

    # Make input tensors
    query = torch.empty(batch_size, num_query_heads, head_dim, dtype=dtype)
    query.uniform_(-scale, scale)

    seq_lens_lst = [random.randint(1, seq_len) for _ in range(batch_size)]  # noqa: S311
    seq_lens_lst[-1] = seq_len
    max_seq_len = max(seq_lens_lst)
    seq_lens = torch.as_tensor(seq_lens_lst, dtype=torch.int)

    # Create the block tables.
    max_num_blocks_per_seq = (max_seq_len + cache_block_size - 1) // cache_block_size
    block_table_lst: list[list[int]] = []
    for _ in range(batch_size):
        block_table = [random.randint(0, num_cache_blocks - 1) for _ in range(max_num_blocks_per_seq)]  # noqa: S311
        block_table_lst.append(block_table)

    block_table = torch.as_tensor(block_table_lst, dtype=torch.int)

    # Create the KV caches.
    key_caches_vllm, value_caches_vllm = create_kv_caches_with_random(
        num_cache_blocks,
        cache_block_size,
        1,
        num_kv_heads,
        head_dim,
        kv_cache_dtype,
        dtype,
        seed,
        device,  # type: ignore[arg-type]
    )

    key_cache_vllm, value_cache_vllm = key_caches_vllm[0], value_caches_vllm[0]

    key_cache_flash, value_cache_flash = reshape_vllm_kvcache(key_cache_vllm, value_cache_vllm)

    return query, key_cache_vllm, value_cache_vllm, key_cache_flash, value_cache_flash, block_table, seq_lens



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Port of vllm rotary_embedding to Triton."""

import torch
import triton
import triton.language as tl


@triton.jit  # type: ignore[misc]
def _rotary_embedding_kernel(
    positions_ptr: tl.const,
    query_ptr: tl.tensor,
    key_ptr: tl.tensor,
    cos_sin_cache_ptr: tl.const,
    rot_dim: int,
    query_stride: int,
    key_stride: int,
    num_heads: int,
    num_kv_heads: int,
    head_size: int,
    cxpr_block_size: tl.constexpr,
) -> None:
    """Apply rotary positional embedding to query and key.

    Args:
        positions_ptr: token positions [num_tokens].
        query_ptr: queries [num_tokens, num_heads * head_size].
        key_ptr: keys [num_tokens, num_heads * head_size].
        cos_sin_cache_ptr: precacluated cos-sin, typically [rot_dim], split into [cos | sin] for NeoX.
        rot_dim: cos_sin_cache_ptr len, typically head_size.
        query_stride: step between token entries.
        key_stride: step between token entries.
        num_heads: number of query heads.
        num_kv_heads: number of kv heads.
        head_size: head size.
        cxpr_block_size: block size (needs to be pow-2).
    """
    # Each block is responsible for one token
    token_idx = tl.program_id(0)
    pos = tl.load(positions_ptr + token_idx)
    # Rotation dep. on token pos; offset to entries for this position
    rot_cache_ptr = cos_sin_cache_ptr + pos * rot_dim

    # "apply_rotary_embedding" device function
    embed_dim = rot_dim // 2
    # For NeoX models, the cos-sin cache is expected to be [cos | sin]
    # Rotation is applied to pairs of entires, so there will be d/2 cos and d/2 sin entries, e.g. one pair of
    # cos-sin values for every two entries
    cos_ptr = rot_cache_ptr
    sin_ptr = rot_cache_ptr + embed_dim

    # Applying rotation to each vector
    # Each "thread" will:
    # - Determine which head its operating on and within this head which pair of entires
    # - Knowing which pair of entries, find the corresponding cos and sin values
    # - Load the entries pair, apply rotation, write back
    #
    # Each token's entry has dimension [num_heads * head_size]
    # The loop ensures that the block of threads will be repeatedly used to cover the entire token entry

    nq = num_heads * embed_dim
    # Offset to query entries for this token
    query_token_offset = token_idx.to(tl.int64) * query_stride
    i = tl.arange(0, cxpr_block_size)
    for _ in tl.range(0, nq, cxpr_block_size):
        head_idx = i // embed_dim
        token_head = query_token_offset + head_idx * head_size
        rot_offset = i % embed_dim

        # "apply_token_rotary_embedding" device function + IS_NEOX=true
        x_index = rot_offset
        y_index = rot_offset + embed_dim
        query_offset = query_ptr + token_head
        mask = i < nq
        x = tl.load(query_offset + x_index, mask=mask)
        y = tl.load(query_offset + y_index, mask=mask)
        cos = tl.load(cos_ptr + x_index, mask=mask)
        sin = tl.load(sin_ptr + x_index, mask=mask)
        x_rot = x * cos - y * sin
        y_rot = y * cos + x * sin
        tl.store(query_offset + x_index, x_rot, mask=mask)
        tl.store(query_offset + y_index, y_rot, mask=mask)
        # Next set of "block IDs"
        i += cxpr_block_size

    nk = num_kv_heads * embed_dim
    # Offset to key entires for this token
    key_token_offset = token_idx.to(tl.int64) * key_stride
    j = tl.arange(0, cxpr_block_size)
    for _ in tl.range(0, nk, cxpr_block_size):
        head_idx = j // embed_dim
        token_head = key_token_offset + head_idx * head_size
        rot_offset = j % embed_dim

        # "apply_token_rotary_embedding" device function + IS_NEOX=true
        x_index = rot_offset
        y_index = rot_offset + embed_dim
        key_offset = key_ptr + token_head
        mask = j < nk
        x = tl.load(key_offset + x_index, mask=mask)
        y = tl.load(key_offset + y_index, mask=mask)
        cos = tl.load(cos_ptr + x_index, mask=mask)
        sin = tl.load(sin_ptr + x_index, mask=mask)
        x_rot = x * cos - y * sin
        y_rot = y * cos + x * sin
        tl.store(key_offset + x_index, x_rot, mask=mask)
        tl.store(key_offset + y_index, y_rot, mask=mask)
        # Next set of "block IDs"
        j += cxpr_block_size


def rotary_embedding_launcher(
    positions: torch.Tensor,
    query: torch.Tensor,
    key: torch.Tensor,
    head_size: int,
    cos_sin_cache: torch.Tensor,
    *,
    is_neox: bool = True,
) -> None:
    """Apply rotary positional embedding to query and key.

    NOTE: Only supporting NeoX model.

    Args:
        positions: token positions [num_tokens].
        query: queries [num_tokens, num_heads * head_size].
        key: keys [num_tokens, num_heads * head_size].
        head_size: head size.
        cos_sin_cache: precacluated cos-sin, typically [rot_dim], split into [cos | sin] for NeoX; use tensor provided by vllm::RotaryEmbedding.
        is_neox: whether using NeoX, must be True; exists mostly for visibility.
    """
    assert is_neox  # noqa: S101
    # Some checks to indirectly ensure that we're not batching
    assert len(positions.shape) == 1 and len(query.shape) == 2 and len(key.shape) == 2  # noqa: S101, PT018, PLR2004

    # Note about using negative indices in shape and stride
    # The commonly (?) accepted organization of data appears to be:
    # - batch, token, other higher level dimensions in the leading positions
    # - data-specific dimensions in the lower positions
    # For example: [batch_size, num_token, embedding_size]
    # Some higher level dimensions may or may not exist.  For example, for unbatched data, there wouldn't be a
    # batch_size dimension.  On the other hand, the lower-level dimensions will always exist.  This is why it makes
    # some sense to count backwards when trying to access dimension information.

    # The use of negative indices may be excessive here given we're enforcing batch==1 (and thus shape), but it's worth
    # doing so to more easily support batched processing in the future.

    num_tokens = query.numel() // query.shape[-1]
    rot_dim = cos_sin_cache.shape[-1]
    num_heads = query.shape[-1] // head_size
    num_kv_heads = key.shape[-1] // head_size
    query_stride = query.stride(-2)
    key_stride = key.stride(-2)

    grid = (num_tokens,)
    block = triton.next_power_of_2(head_size)
    _rotary_embedding_kernel[grid](
        positions,
        query,
        key,
        cos_sin_cache,
        rot_dim,
        query_stride,
        key_stride,
        num_heads,
        num_kv_heads,
        head_size,
        block,
    )
